import hashlib
import math
import os
import re
from email import policy
from email.message import Message as PyEmailMessage
from email.parser import BytesParser
from email.utils import parsedate_to_datetime
from typing import Any, Dict, List, Optional, Tuple

from django.conf import settings
from django.contrib import admin
from django.core.cache import cache
from django.core.mail import EmailMessage
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import render
from django.utils.html import conditional_escape
from django.utils.translation import gettext as _
from django.views.decorators.clickjacking import xframe_options_exempt

# Used to split aggregated messages stored in a single raw file (bytes)
AGGREGATED_EMAIL_SPLIT_RE = re.compile(rb"\r?\n-{50,}[ \t]*\r?\n")


def _ensure_supported_backend() -> None:
    if settings.EMAIL_BACKEND not in (
        "django.core.mail.backends.locmem.EmailBackend",
        "django.core.mail.backends.filebased.EmailBackend",
    ):
        raise Http404()


def _is_locmem_backend() -> bool:
    return settings.EMAIL_BACKEND == "django.core.mail.backends.locmem.EmailBackend"


def _is_filebased_backend() -> bool:
    return settings.EMAIL_BACKEND == "django.core.mail.backends.filebased.EmailBackend"


def _get_email_file_basedir() -> Optional[str]:
    basedir = getattr(settings, "EMAIL_FILE_PATH", None)
    if not basedir or not os.path.isdir(basedir):
        return None
    return basedir


def _fetch_outbox_locmem() -> List[EmailMessage]:
    from django.core import mail  # Imported lazily to ensure correct backend

    outbox: List[EmailMessage] = getattr(mail, "outbox", None)
    messages: List[EmailMessage] = []
    for em in outbox or []:
        _populate_metadata_from_locmem(em)
        messages.append(em)
    return messages


def _collect_file_paths_sorted(basedir: str) -> List[str]:
    filepaths: List[str] = [
        os.path.join(basedir, name)
        for name in os.listdir(basedir)
        if os.path.isfile(os.path.join(basedir, name))
    ]
    filepaths.sort(key=lambda p: os.path.getmtime(p))
    return filepaths


def _read_file_bytes(path: str) -> Optional[bytes]:
    try:
        with open(path, "rb") as fh:
            return fh.read()
    except Exception:
        return None


def _split_aggregated_messages(raw_bytes: bytes) -> List[bytes]:
    try:
        parts = AGGREGATED_EMAIL_SPLIT_RE.split(raw_bytes)
        return parts if len(parts) > 1 else [raw_bytes]
    except Exception:
        return [raw_bytes]


def _parse_raw_to_messages(raw_bytes: bytes) -> List[PyEmailMessage]:
    parsed: List[PyEmailMessage] = []
    for chunk in _split_aggregated_messages(raw_bytes):
        if not chunk or chunk.strip() == b"":
            continue
        try:
            parsed.append(BytesParser(policy=policy.default).parsebytes(chunk))
        except Exception:
            # Ignore chunks that fail to parse; we'll fall back to whole payload if needed
            pass
    return parsed


def _build_django_email_from_parsed(
    msg: PyEmailMessage, path: Optional[str]
) -> EmailMessage:
    subject = str(msg.get("Subject", ""))
    from_email = str(msg.get("From", ""))
    to = msg.get_all("To", []) or []
    cc = msg.get_all("Cc", []) or []
    bcc = msg.get_all("Bcc", []) or []
    sent_at = _parse_date_header(msg.get("Date"))
    if sent_at is None and path:
        try:
            sent_at = _safe_dt(os.path.getmtime(path))
        except Exception:
            sent_at = None

    em = EmailMessage(subject=subject, body="", from_email=from_email, to=to)
    if cc:
        em.cc = cc
    if bcc:
        em.bcc = bcc
    em.sent_at = sent_at

    if msg.is_multipart():
        alts = []
        for part in msg.walk():
            ctype = part.get_content_type()
            if ctype == "text/html":
                try:
                    alts.append((part.get_content(), "text/html"))
                except Exception:
                    try:
                        content = part.get_payload(decode=True) or b""
                        alts.append((content.decode("utf-8", "replace"), "text/html"))
                    except Exception:
                        pass
            elif ctype == "text/plain" and not em.body:
                try:
                    em.body = part.get_content()
                except Exception:
                    try:
                        content = part.get_payload(decode=True) or b""
                        em.body = content.decode("utf-8", "replace")
                    except Exception:
                        em.body = ""
        _populate_attachments_from_parsed(msg, em)
        if alts:
            em.alternatives = alts
    else:
        ctype = msg.get_content_type()
        try:
            content_value = msg.get_content()
        except Exception:
            try:
                content_value = (msg.get_payload(decode=True) or b"").decode(
                    "utf-8", "replace"
                )
            except Exception:
                content_value = ""
        if ctype == "text/html":
            em.body = ""
            em.alternatives = [(content_value, "text/html")]
        else:
            em.body = content_value
        _populate_attachments_from_parsed(msg, em)

    return em


def _fetch_outbox_filebased() -> List[EmailMessage]:
    basedir = _get_email_file_basedir()
    if not basedir:
        return []
    messages: List[EmailMessage] = []
    for path in _collect_file_paths_sorted(basedir):
        raw_bytes = _read_file_bytes(path)
        if raw_bytes is None:
            continue
        for parsed in _parse_raw_to_messages(raw_bytes):
            em = _build_django_email_from_parsed(parsed, path)
            messages.append(em)
    return messages


def _filebased_cache_key(basedir: str) -> str:
    return (
        "email_outbox:filebased:index:"
        + hashlib.sha256(basedir.encode("utf-8", "surrogatepass")).hexdigest()
    )


def _compute_filebased_fingerprint(basedir: str) -> str:
    try:
        entries: List[Tuple[str, float, int]] = []
        for name in os.listdir(basedir):
            path = os.path.join(basedir, name)
            if os.path.isfile(path):
                try:
                    entries.append(
                        (name, os.path.getmtime(path), os.path.getsize(path))
                    )
                except Exception:
                    continue
        entries.sort(key=lambda t: t[0])
        data = repr(entries).encode("utf-8", "surrogatepass")
        return hashlib.sha256(data).hexdigest()
    except Exception:
        return "0"


def _get_filebased_index_cached() -> List[Tuple[str, int]]:
    basedir = _get_email_file_basedir()
    if not basedir:
        return []
    key = _filebased_cache_key(basedir)
    current_fp = _compute_filebased_fingerprint(basedir)
    cached = cache.get(key)
    if isinstance(cached, dict) and cached.get("fingerprint") == current_fp:
        index = cached.get("index") or []
        if isinstance(index, list):
            return index

    # Rebuild index
    index: List[Tuple[str, int]] = []
    for path in _collect_file_paths_sorted(basedir):
        raw = _read_file_bytes(path)
        if raw is None:
            continue
        parts = _split_aggregated_messages(raw)
        parts = [p for p in parts if p and p.strip() != b""]
        if not parts:
            parts = [raw]
        for chunk_idx, unused_part in enumerate(parts):
            index.append((path, chunk_idx))

    cache.set(key, {"fingerprint": current_fp, "index": index}, 300)
    return index


def _refresh_filebased_index() -> List[Tuple[str, int]]:
    basedir = _get_email_file_basedir()
    if not basedir:
        return []
    key = _filebased_cache_key(basedir)
    cache.delete(key)
    return _get_filebased_index_cached()


def _get_message_filebased_by_idx(idx: int) -> EmailMessage:
    basedir = _get_email_file_basedir()
    if not basedir:
        raise Http404()

    def _load_with_index() -> Tuple[str, int, bytes, List[bytes]]:
        index = _get_filebased_index_cached()
        try:
            path, chunk_idx = index[idx]
        except (IndexError, TypeError):
            raise Http404()
        raw = _read_file_bytes(path)
        if raw is None:
            raise FileNotFoundError()
        parts = _split_aggregated_messages(raw)
        parts = [p for p in parts if p and p.strip() != b""]
        if not parts:
            parts = [raw]
        return path, chunk_idx, raw, parts

    try:
        path, chunk_idx, raw, parts = _load_with_index()
        if chunk_idx >= len(parts):
            raise IndexError()
    except (FileNotFoundError, IndexError):
        # Likely stale cache; refresh and retry once
        _refresh_filebased_index()
        path, chunk_idx, raw, parts = _load_with_index()
        if chunk_idx >= len(parts):
            raise Http404()

    chunk = parts[chunk_idx]
    try:
        py_msg = BytesParser(policy=policy.default).parsebytes(chunk)
    except Exception:
        # Fallback to parsing the whole file if specific chunk fails
        try:
            py_msg = BytesParser(policy=policy.default).parsebytes(raw)
        except Exception:
            raise Http404()
    return _build_django_email_from_parsed(py_msg, path)


def _get_outbox() -> List[EmailMessage]:
    if _is_locmem_backend():
        return _fetch_outbox_locmem()
    if _is_filebased_backend():
        return _fetch_outbox_filebased()
    return []


def _get_message(idx: int) -> EmailMessage:
    if _is_filebased_backend():
        try:
            return _get_message_filebased_by_idx(idx)
        except Http404:
            raise
        except Exception:
            # Fallback to full scan if optimized path fails
            pass
    outbox = _get_outbox()
    try:
        return outbox[idx]
    except (IndexError, TypeError):
        raise Http404()


def outbox_list_view(request: HttpRequest) -> HttpResponse:
    _ensure_supported_backend()
    emails = _get_outbox()
    # Order by descending original index (most recently added first)
    emails_sorted = sorted(enumerate(emails), key=lambda im: im[0], reverse=True)

    # Pagination
    try:
        page = int(request.GET.get("p", "1"))
    except ValueError:
        page = 1
    try:
        per_page = int(request.GET.get("per_page", "50"))
    except ValueError:
        per_page = 50
    count = len(emails_sorted)
    num_pages = max(1, math.ceil(count / per_page))
    if page < 1:
        page = 1
    if page > num_pages:
        page = num_pages
    start = (page - 1) * per_page
    end = start + per_page
    page_items = emails_sorted[start:end]

    # Build a view-model with safe defaults for template
    vm = [
        {
            "idx": orig_idx,
            "subject": m.subject or "",
            "from_email": m.from_email or "",
            "to": list(m.to or []),
            "cc": list(getattr(m, "cc", []) or []),
            "bcc": list(getattr(m, "bcc", []) or []),
            "sent_at": getattr(m, "sent_at", None),
            "attachments_count": len(getattr(m, "attachments_meta", []) or []),
        }
        for (orig_idx, m) in page_items
    ]

    # Compact page range around current page
    pr_start = max(1, page - 5)
    pr_end = min(num_pages, page + 5)
    page_range = list(range(pr_start, pr_end + 1))

    context = {
        "title": _("Emails"),
        "emails": vm,
        "page": page,
        "num_pages": num_pages,
        "has_prev": page > 1,
        "has_next": page < num_pages,
        "prev_page": page - 1,
        "next_page": page + 1,
        "page_range": page_range,
        "count": count,
        "per_page": per_page,
    }
    context.update(admin.site.each_context(request))
    return render(request, "admin/outbox_list.html", context)


def outbox_detail_view(request: HttpRequest, idx: str) -> HttpResponse:
    _ensure_supported_backend()
    message = _get_message(int(idx))
    context = {
        "title": _("Email #%(num)s") % {"num": idx},
        "idx": int(idx),
        "subject": message.subject or "",
        "from_email": message.from_email or "",
        "to": list(message.to or []),
        "cc": list(getattr(message, "cc", []) or []),
        "bcc": list(getattr(message, "bcc", []) or []),
        "headers": getattr(message, "extra_headers", {}) or {},
        "sent_at": getattr(message, "sent_at", None),
        "attachments": [
            {"aidx": i, **meta}
            for i, meta in enumerate(getattr(message, "attachments_meta", []) or [])
        ],
    }
    context.update(admin.site.each_context(request))
    return render(request, "admin/outbox_detail.html", context)


def _extract_html_body(message: EmailMessage) -> Tuple[str, str]:
    # Returns (body_html, content_type)
    # Prefer HTML alternative if present (EmailMultiAlternatives)
    alternatives = getattr(message, "alternatives", None)
    if alternatives:
        for content, mime in alternatives:
            if mime == "text/html":
                return _ensure_light_styles(content), "text/html"
    # Fallback to plain text body
    text_body = message.body or ""
    # Some backends/log exports add a trailing separator line of many dashes
    escaped = conditional_escape(text_body).replace("\n", "<br>")
    html = (
        '<!DOCTYPE html><html><head><meta charset="utf-8"></head>'
        "<body>{}</body></html>".format(escaped)
    )
    html = _ensure_light_styles(html)
    return html, "text/html"


def _ensure_light_styles(html: str) -> str:
    """Inject minimal CSS to force a light background and readable text.

    Attempts to add a <style> block either inside <head>, or right before <body>,
    or wraps the content if no standard structure is found.
    """
    style = (
        "<style>html,body{background:#fff !important;color:#000 !important;}"
        "a{color:#0645ad}img{max-width:100%;height:auto}</style>"
    )

    # Insert into <head> if present
    if re.search(r"<head[^>]*>", html, flags=re.IGNORECASE):
        return re.sub(
            r"(<head[^>]*>)", r"\1" + style, html, count=1, flags=re.IGNORECASE
        )

    # Else, insert immediately before <body>
    if re.search(r"<body[^>]*>", html, flags=re.IGNORECASE):
        return re.sub(
            r"(<body[^>]*>)", style + r"\1", html, count=1, flags=re.IGNORECASE
        )

    # Else, wrap content
    return (
        '<!DOCTYPE html><html><head><meta charset="utf-8">' + style + "</head>"
        "<body>" + html + "</body></html>"
    )


@xframe_options_exempt
def outbox_body_view(request: HttpRequest, idx: str) -> HttpResponse:
    _ensure_supported_backend()
    message = _get_message(int(idx))
    body, content_type = _extract_html_body(message)
    return HttpResponse(body, content_type=content_type)


def outbox_attachment_view(request: HttpRequest, idx: str, aidx: str) -> HttpResponse:
    _ensure_supported_backend()
    message = _get_message(int(idx))
    attachments: List[Dict[str, Any]] = getattr(message, "attachments_data", []) or []
    try:
        att = attachments[int(aidx)]
    except (IndexError, ValueError, TypeError):
        raise Http404()
    filename = att.get("name") or "attachment"
    content = att.get("content") or b""
    content_type = att.get("content_type") or "application/octet-stream"
    response = HttpResponse(content, content_type=content_type)
    response["Content-Disposition"] = 'attachment; filename="%s"' % filename
    return response


def _populate_metadata_from_locmem(em: EmailMessage) -> None:
    # Attempt to extract Date header from generated message
    try:
        py_msg: PyEmailMessage = em.message()
        sent_at = _parse_date_header(py_msg.get("Date"))
    except Exception:
        sent_at = None
    em.sent_at = sent_at
    _populate_attachments_from_locmem(em)


def _populate_attachments_from_locmem(em: EmailMessage) -> None:
    meta: List[Dict[str, Any]] = []
    data: List[Dict[str, Any]] = []
    for att in getattr(em, "attachments", []) or []:
        if isinstance(att, tuple) and len(att) in (2, 3):
            # (filename, content[, mimetype])
            name = att[0] or "attachment"
            content = att[1]
            if isinstance(content, str):
                content_bytes = content.encode("utf-8")
            else:
                content_bytes = content or b""
            ctype = att[2] if len(att) == 3 else "application/octet-stream"
            meta.append(
                {"name": name, "size": len(content_bytes), "content_type": ctype}
            )
            data.append({"name": name, "content": content_bytes, "content_type": ctype})
        else:
            # MIMEBase-like
            try:
                name = getattr(att, "get_filename", lambda: None)() or "attachment"
                ctype = getattr(
                    att, "get_content_type", lambda: "application/octet-stream"
                )()
                payload = getattr(att, "get_payload", lambda decode=True: b"")(
                    decode=True
                )
                content_bytes = payload or b""
                meta.append(
                    {"name": name, "size": len(content_bytes), "content_type": ctype}
                )
                data.append(
                    {"name": name, "content": content_bytes, "content_type": ctype}
                )
            except Exception:
                continue
    em.attachments_meta = meta
    em.attachments_data = data


def _populate_attachments_from_parsed(msg: PyEmailMessage, em: EmailMessage) -> None:
    meta: List[Dict[str, Any]] = []
    data: List[Dict[str, Any]] = []
    if msg.is_multipart():
        for part in msg.walk():
            disp = (part.get("Content-Disposition") or "").lower()
            if "attachment" in disp or (part.get_filename()):
                name = part.get_filename() or "attachment"
                ctype = part.get_content_type() or "application/octet-stream"
                try:
                    content_bytes = part.get_content()
                    if isinstance(content_bytes, str):
                        content_bytes = content_bytes.encode("utf-8")
                except Exception:
                    try:
                        content_bytes = part.get_payload(decode=True) or b""
                    except Exception:
                        content_bytes = b""
                meta.append(
                    {"name": name, "size": len(content_bytes), "content_type": ctype}
                )
                data.append(
                    {"name": name, "content": content_bytes, "content_type": ctype}
                )
    else:
        # Single-part message isn't an attachment container
        pass
    em.attachments_meta = meta
    em.attachments_data = data


def _parse_date_header(date_value: Optional[str]):
    if not date_value:
        return None
    try:
        return parsedate_to_datetime(str(date_value))
    except Exception:
        return None


def _safe_dt(timestamp: float):
    try:
        import datetime as _dt

        return _dt.datetime.fromtimestamp(timestamp)
    except Exception:
        return None



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Sylvain Fankhauser <sylvain.fankhauser@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from decimal import ROUND_HALF_EVEN
from decimal import Decimal as D
from typing import Iterable

from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import F, Sum
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy as _p

from simple_history.utils import get_history_manager_for_model

from apps.orga.models import Organization

from .season import Season
from .session import Session
from .settings import AnnualStateSetting

HistoricalSession = get_history_manager_for_model(Session).model


class Invoice(models.Model):
    STATUS_DRAFT = 0
    STATUS_VALIDATED = 1
    STATUS_CHOICES = (
        (STATUS_DRAFT, _("Brouillon")),
        (STATUS_VALIDATED, _("Validée")),
    )

    generated_at = models.DateTimeField(default=timezone.now)
    status = models.IntegerField(
        _("Statut"), choices=STATUS_CHOICES, default=STATUS_DRAFT
    )
    ref = models.CharField(_("Référence"), max_length=20, blank=False, unique=True)
    title = models.CharField(_("Titre"), max_length=255, blank=True)
    organization = models.ForeignKey(
        Organization,
        verbose_name=_("Établissement"),
        on_delete=models.PROTECT,
        related_name="invoices",
    )
    season = models.ForeignKey(
        Season,
        verbose_name=_p("Singular month", "Mois"),
        on_delete=models.PROTECT,
        related_name="invoices",
    )

    def __str__(self):
        return gettext(f"Facture {self.ref} pour {self.organization} / {self.season}")

    def get_absolute_url(self):
        return reverse_lazy(
            "invoice-detail",
            kwargs={
                "seasonpk": self.season.pk,
                "orgapk": self.organization.pk,
                "invoiceref": self.ref,
            },
        )

    @property
    def sessions(self):
        return self.season.sessions.filter(orga=self.organization)

    def sum_of(self, things: Iterable[str]):
        if self.pk:
            fields = sum(F(thing) for thing in things)
            return self.lines.aggregate(total=Sum(fields))["total"]
        return None

    def sum_cost_bikes_reduced(self):
        return sum([line.cost_bikes_reduced for line in self.lines.all()])

    def sum_cost(self):
        return self.sum_cost_bikes_reduced() + (self.sum_cost_participants or 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Generate sum_* methods
        for thing in [
            "nb_bikes",
            "nb_participants",
            "cost_participants",
        ]:
            setattr(self, f"sum_{thing}", self.sum_of([thing]))

    @property
    def is_locked(self):
        return self.status == self.STATUS_VALIDATED

    @property
    def status_class(self):
        css_class = "default"
        if self.status == self.STATUS_DRAFT:
            css_class = "warning"  # Orange
        elif self.status == self.STATUS_VALIDATED:
            css_class = "success"  # Green
        return css_class

    @cached_property
    def settings(self):
        try:
            return AnnualStateSetting.objects.get(
                canton=self.organization.address_canton,
                year=self.season.year,
            )
        except AnnualStateSetting.DoesNotExist:
            return AnnualStateSetting()

    @property
    def is_up_to_date(self):
        """
        Check if the whole Invoice is up_to_date
        """
        # First check if the individual lines are OK.
        lines = self.lines.prefetch_related("session", "historical_session")
        if not all([line.is_up_to_date for line in lines.all()]):
            return False
        # Check if the invoice lines correspond to the concerned sessions
        if set(lines.values_list("session_id", flat=True)) != set(
            self.sessions.values_list("id", flat=True)
        ):
            return False
        return True

    @property
    def adjusted_sum_of_bikes(self):
        lines_per_day = dict()
        for line in self.lines.all():
            date_of_current_line = line.historical_session.day
            nb_bikes = lines_per_day.get(date_of_current_line, {}).get(
                "max_nb_bikes", 0
            )
            lines_per_day[date_of_current_line] = {
                "max_nb_bikes": max(line.nb_bikes, nb_bikes)
            }

        adjusted_sum_of_bikes = sum(
            [line["max_nb_bikes"] for _, line in lines_per_day.items()]
        )

        return adjusted_sum_of_bikes

    @property
    def month_of_the_invoice(self):
        return self.generated_at.strftime("%B")


def round_CHF(n: D):
    """
    Round a Decimal to 0.05 cents
    """
    return D((n * 20).quantize(D("1"), rounding=ROUND_HALF_EVEN) / 20).quantize(
        D("0.01")
    )


class InvoiceLine(models.Model):
    session = models.ForeignKey(Session, on_delete=models.SET_NULL, null=True)
    historical_session = models.ForeignKey(HistoricalSession, on_delete=models.PROTECT)
    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE, related_name="lines")
    nb_bikes = models.PositiveSmallIntegerField()
    nb_participants = models.PositiveSmallIntegerField()
    cost_bikes = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        validators=[MinValueValidator(0)],
    )
    cost_participants = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        validators=[MinValueValidator(0)],
    )

    class Meta:
        unique_together = (("session", "invoice"),)
        ordering = ("historical_session__day", "historical_session__begin")

    def __str__(self):
        return gettext(
            f"{self.invoice.ref}: {self.historical_session} - Vélos: {self.nb_bikes} ({self.cost_bikes} CHF) - Participants: {self.nb_participants} ({self.cost_participants} CHF)"
        )

    @property
    def has_cost_bikes(self):
        """
        True when this line is the one accounting for the cost of bikes.
        There may be multiple lines (sessions) on the same day,
        but bikes are rented for the whole day.
        """
        last_line_with_most_bikes = (
            self.invoice.lines.filter(
                historical_session__day=self.historical_session.day
            )
            .order_by("-nb_bikes", "-historical_session__begin", "-id")
            .first()
        )
        return self == last_line_with_most_bikes

    @property
    def cost_bikes_reduced(self):
        if not self.has_cost_bikes:
            return 0
        # Sum of the costs, rounded to 5 cents.
        return round_CHF(
            D(self.cost_bikes * D(1 - self.cost_bikes_reduction_percent() / 100))
        )

    @property
    def cost(self):
        return self.cost_bikes_reduced + self.cost_participants

    def most_recent_historical_session(self):
        """
        Provide a shortcut to getting the latest historical copy of the session
        """
        try:
            return self.session.history.latest("history_date")
        except AttributeError:  # Without a self.session (deleted ?)
            return None

    def save(self, *args, **kwargs):
        """
        Safeguard against not having a historical_session
        """
        if not self.historical_session_id:
            self.historical_session = self.most_recent_historical_session()
        super().save(*args, **kwargs)

    @cached_property
    def is_up_to_date(self):
        """
        Check if that invoice line is up-to-date by comparing the historical session, number and costs
        """
        if not self.session:
            return False

        if (
            self.historical_session.history_id
            != self.most_recent_historical_session().history_id
        ):
            return False

        for t in ["bike", "participant"]:
            # Check if the nb_* is identical
            if getattr(self, f"nb_{t}s") != getattr(self.session, f"n_{t}s"):
                return False
            # Check if the calculated cost is identical
            if getattr(self, f"cost_{t}s") != getattr(
                self.session, f"n_{t}s"
            ) * getattr(self.invoice.settings, f"cost_per_{t}"):
                return False
        return True

    def cost_bikes_reduction_percent(self):
        """
        Compute the reduction in cost for bikes. Reductions are applied for consecutive days.
        """
        # Maps "consecutive days" to "percentage of reduction"
        reductions_in_percent = {2: 5, 3: 10, 4: 20, 5: 30}
        # Get the other lines of this invoice, with their daily differences to ours
        lines_diff_days = [
            diff_days.days
            for diff_days in self.invoice.lines.annotate(
                diff_days=(F("historical_session__day") - self.historical_session.day)
            ).values_list("diff_days", flat=True)
        ]

        max_days_diff_to_check = max(reductions_in_percent.keys())

        # Check the days before first
        nth_in_serie = 0
        while nth_in_serie < max_days_diff_to_check:
            if -nth_in_serie - 1 not in lines_diff_days:
                break
            nth_in_serie = nth_in_serie + 1
        # Check the days after
        followed_by = 0
        while followed_by < max_days_diff_to_check:
            if followed_by + 1 not in lines_diff_days:
                break
            followed_by = followed_by + 1
        # We now know we're in a sequence of that many lines:
        in_a_sequence_of = nth_in_serie + 1 + followed_by
        # Get me the reduction
        return reductions_in_percent.get(in_a_sequence_of, 0)

    def refresh(self):
        """
        Align the invoiceline's attributes
        """
        if self.session:
            self.historical_session = self.most_recent_historical_session()
            self.nb_bikes = self.session.n_bikes
            self.nb_participants = self.session.n_participants
            self.cost_bikes = self.invoice.settings.cost_per_bike * self.session.n_bikes
            self.cost_participants = (
                self.invoice.settings.cost_per_participant * self.session.n_participants
            )
        # Clear the cached is_up_to_date if possible
        try:
            del self.is_up_to_date
        except AttributeError:
            pass



# Generated by Django 2.2.9 on 2020-03-04 08:31

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Timesheet",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateField(blank=True, null=True, verbose_name="Date")),
                (
                    "time_monitor",
                    models.FloatField(
                        blank=True, null=True, verbose_name="Heure moniteur"
                    ),
                ),
                (
                    "time_actor",
                    models.FloatField(
                        blank=True, null=True, verbose_name="Heure intervenant"
                    ),
                ),
                (
                    "overtime",
                    models.FloatField(default=0, verbose_name="Heures supplémentaires"),
                ),
                (
                    "traveltime",
                    models.FloatField(default=0, verbose_name="Heures de trajet"),
                ),
                ("validated_at", models.DateTimeField(blank=True, null=True)),
                ("comments", models.TextField(blank=True, verbose_name="Remarques")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="timesheets",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "date")},
            },
        ),
    ]



# Generated by Django 4.2.6 on 2025-12-01 12:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('salary', '0018_add_states_ow_nw'),
    ]

    operations = [
        migrations.AlterField(
            model_name='timesheet',
            name='leader_count',
            field=models.IntegerField(default=0, verbose_name='Participation(s) comme moniteur·trice 2'),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from django.utils.translation import activate

from allauth.account.models import EmailAddress

User = get_user_model()


class HomePageRedirectTest(TestCase):
    def setUp(self):
        self.userpassword = "top_secret"
        self.user = User.objects.create_user(
            username="jacob", email="jacob@example.com", password=self.userpassword
        )
        self.language = "fr"
        activate(self.language)

    def test_unauthenticated_homepage_redirects(self):
        # Issue a GET request.
        response = self.client.get("/")
        i18n_homepage = "/%s/" % self.language
        # Check that the response is 302 redirect to the language page
        self.assertRedirects(response, i18n_homepage, target_status_code=302)

        # Now fetch that language-based homepage
        response = self.client.get(i18n_homepage)

        # Check that the response is 302 redirect to the login page
        self.assertRedirects(
            response,
            "{login}?next={home}".format(
                login=reverse("account_login"), home=i18n_homepage
            ),
        )

    def test_login_with_username_fails(self):
        response = self.client.post(
            reverse("account_login"),
            {
                "login": self.user.username,
                "password": self.userpassword,
            },
        )
        # The login failed, we're back on the same page
        self.assertTemplateUsed(response, "account/login.html")
        self.assertEqual(response.status_code, 200)

    def test_login_process_with_uncertified_email(self):
        response = self.client.post(
            reverse("account_login"),
            {
                "login": self.user.email,
                "password": self.userpassword,
            },
        )
        # This triggered a mail sending to confirm that email
        self.assertEqual(EmailAddress.objects.count(), 1)
        self.assertFalse(EmailAddress.objects.first().verified)

        # Now confirm that email
        ec = EmailAddress.objects.get(user=self.user)
        ec.verified = True
        ec.set_as_primary(conditional=True)
        ec.save()
        # And retry login:
        response = self.client.post(
            reverse("account_login"),
            {
                "login": self.user.email,
                "password": self.userpassword,
            },
        )
        # This time, login worked
        self.assertTemplateUsed(response, "account/messages/logged_in.txt")
        self.assertEqual(response.status_code, 302)

    def test_signup_forbidden(self):
        response = self.client.get(reverse("account_signup"))

        # Check that the response is a 200 that uses the forbidden
        # signup template
        self.assertTemplateUsed(response, "account/signup_closed.html")
        self.assertEqual(response.status_code, 200)



# Generated by Django 1.11.4 on 2017-09-05 13:16
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0046_add_helpersessionavailability_chosen_as"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="helpersessionavailability",
            name="chosen",
        ),
    ]



# Generated by Django 1.11.5 on 2017-09-07 10:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0047_remove_helpersessionavailability_chosen"),
    ]

    operations = [
        migrations.AlterField(
            model_name="helpersessionavailability",
            name="chosen_as",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (0, "Pas choisi"),
                    (1, "Choisi"),
                    (2, "Comme intervenant"),
                    (5, "Moniteur de Secours"),
                    (3, "Moniteur 1"),
                    (4, "Moniteur 2"),
                ],
                default=0,
                verbose_name="S\xe9lectionn\xe9 pour la session comme",
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from django.test import TestCase
from django.utils.html import escape

from defivelo.templatetags.dv_filters import tel_link


class TelLinkTestCase(TestCase):
    """Tests for the tel_link template filter"""

    def test_empty_string(self):
        """Test that empty string returns empty string"""
        result = tel_link("")
        self.assertEqual(result, "")

    def test_none_value(self):
        """Test that None returns empty string"""
        result = tel_link(None)
        self.assertEqual(result, "")

    def test_swiss_international_format(self):
        """Test Swiss international format +41XXXXXXXXXX"""
        result = tel_link("+41791234567")
        self.assertEqual('<a href="tel:+41791234567">079 123 45 67</a>', result)

    def test_swiss_local_format(self):
        """Test Swiss local format starting with 0"""
        result = tel_link("079 123 45 67")
        self.assertEqual('<a href="tel:+41791234567">079 123 45 67</a>', result)

    def test_swiss_local_format_no_spaces(self):
        """Test Swiss local format without spaces"""
        result = tel_link("0791234567")

        # Should be formatted
        self.assertEqual('<a href="tel:+41791234567">079 123 45 67</a>', result)

    def test_international_non_swiss(self):
        """Test international format for non-Swiss number"""
        # French number
        result = tel_link("+33123456789")

        # Should contain the tel: link with international format
        self.assertIn('href="tel:+33123456789"', result)

        # Should display in national format (without country code)
        self.assertIn("01 23 45 67 89", result)

        # Should be a link
        self.assertIn("<a ", result)
        self.assertIn("</a>", result)

    def test_invalid_phone_number(self):
        """Test invalid phone number returns as-is in display"""
        result = tel_link("invalid")
        self.assertEqual('<a href="tel:invalid">invalid</a>', result)

    def test_invalid_international(self):
        """Test invalid international"""
        result = tel_link("+4179123")
        self.assertEqual('<a href="tel:+4179123">79123</a>', result)

    def test_html_safety(self):
        """Test that the result is marked as safe HTML"""
        result = tel_link("+41791234567")

        # The result should be a SafeString (marked safe)
        from django.utils.safestring import SafeString

        self.assertIsInstance(result, SafeString)

    def test_unsafe_number(self):
        bad_input = "<script>alert('hack')</script>"
        result = tel_link(bad_input)

        # Should escape the link
        self.assertIn(f'href="tel:{escape(bad_input)}"', result)



from django import forms
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.forms import formset_factory
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from rolepermissions.checkers import has_role

from apps.challenge.models.session import Session
from apps.common.fields import CheckboxInput, NumberInput, TimeNumberInput
from apps.salary.models import (
    MonthlyCantonalValidation,
    MonthlyCantonalValidationUrl,
    Timesheet,
)

from . import BONUS_LEADER, HOURLY_RATE_HELPER, RATE_ACTOR


class TimesheetFormBase(forms.ModelForm):
    class Meta:
        model = Timesheet
        readonly = ("date", "time_helper", "actor_count", "leader_count", "ignore")
        fields = [
            "date",
            "time_helper",
            "actor_count",
            "leader_count",
            "overtime",
            "traveltime",
            "comments",
            "ignore",
        ]
        labels = {
            "time_helper": format_lazy(
                _("Heures moni·teur·trice ({price}.-/h)"), price=HOURLY_RATE_HELPER
            ),
            "actor_count": format_lazy(
                _("Intervention(s) ({price}.-/Qualif’)"), price=RATE_ACTOR
            ),
            "leader_count": format_lazy(
                _("Participation(s) comme moni·teur·trice 2 ({price}.-/Qualif’)"),
                price=BONUS_LEADER,
            ),
            "overtime": format_lazy(
                _("Heures supplémentaires ({price}.-/h)"), price=HOURLY_RATE_HELPER
            ),
            "traveltime": _("Heures de trajet (aller-retour)"),
            "ignore": _("Ne compter aucune heure de travail"),
        }

        widgets = {
            "date": forms.HiddenInput(),
            "time_helper": TimeNumberInput(
                attrs={
                    "readonly": "readonly",
                    "class": "hide",
                    "data-unit-price": HOURLY_RATE_HELPER,
                },
            ),
            "actor_count": NumberInput(
                attrs={
                    "readonly": "readonly",
                    "class": "hide",
                    "data-unit-price": RATE_ACTOR,
                },
            ),
            "leader_count": NumberInput(
                attrs={
                    "readonly": "readonly",
                    "class": "hide",
                    "data-unit-price": BONUS_LEADER,
                },
            ),
            "overtime": TimeNumberInput(
                attrs={
                    "step": 0.25,
                    "min": -10,
                    "max": 10,
                    "class": "hide",
                    "data-unit-price": HOURLY_RATE_HELPER,
                }
            ),
            "traveltime": TimeNumberInput(
                attrs={
                    "step": 1,
                    "min": 0,
                    "max": 2,
                    "class": "hide",
                    "data-unit-price": HOURLY_RATE_HELPER,
                }
            ),
            "ignore": CheckboxInput(
                attrs={
                    "data-function": "ignore",
                    "readonly": "readonly",
                    "disabled": "disabled",
                }
            ),
            "comments": forms.Textarea(attrs={"rows": 3, "cols": 20}),
        }

    def __init__(self, selected_user, validator, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validator = validator
        self.selected_user = selected_user

    def clean(self):
        cleaned_data = super().clean()
        if not (
            Session.objects.values("day")
            .filter(
                Q(qualifications__actor=self.selected_user)
                | Q(qualifications__helpers=self.selected_user)
                | Q(qualifications__leader=self.selected_user)
            )
            .filter(day=cleaned_data["date"])
            .exists()
        ):
            raise forms.ValidationError(
                _(
                    "Vous ne pouvez pas rentrer des heures pour le %(day)s: aucune qualif ce jour-ci."
                ),
                params={"day": cleaned_data["date"]},
            )

        if cleaned_data["overtime"] != 0.0 and not cleaned_data["comments"]:
            self.add_error(
                "comments",
                forms.ValidationError(
                    _(
                        "Les remarques doivent être renseignées pour tout prétention d’heures supplémentaires"
                    )
                ),
            )

        return cleaned_data

    def save(self):
        if self.errors:
            raise ValueError(
                "The %s could not be %s because the data didn't validate."
                % (
                    self.instance._meta.object_name,
                    "created" if self.instance._state.adding else "changed",
                )
            )
        timesheet, created = Timesheet.objects.update_or_create(
            user=self.selected_user,
            date=self.cleaned_data["date"],
            defaults=self.cleaned_data,
        )
        return timesheet


class TimesheetForm(TimesheetFormBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.initial.get("validated"):
            for key in self.fields:
                self.fields[key].widget.attrs["readonly"] = True
                self.fields[key].disabled = True

        # The "ignore" field is disabled for non-editors
        self.fields["ignore"].widget.attrs["readonly"] = "readonly"
        self.fields["ignore"].widget.attrs["disabled"] = "disabled"

        # Show the "ignore" field only if it was ticked by a StateManager.
        if not self.initial.get("ignore"):
            self.fields["ignore"].widget = forms.HiddenInput()

    def clean(self):
        """
        Never allow ignore to be set through this one form
        """
        cleaned_data = super().clean()
        del cleaned_data["ignore"]
        return cleaned_data


class ControlTimesheetForm(TimesheetFormBase):
    validated = forms.BooleanField(
        label=_("Valider"), required=False, widget=CheckboxInput()
    )

    class Meta(TimesheetFormBase.Meta):
        fields = TimesheetFormBase.Meta.fields + ["validated"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.initial.get("validated") and not has_role(self.validator, "power_user"):
            for key in self.fields:
                self.fields[key].widget.attrs["readonly"] = True
                self.fields[key].disabled = True

        # StateManagers can fiddle with it
        del self.fields["ignore"].widget.attrs["readonly"]
        del self.fields["ignore"].widget.attrs["disabled"]

    def clean(self):
        cleaned_data = super().clean()
        if self.initial["validated"] and not cleaned_data.get("validated"):
            cleaned_data["validated_at"] = None
            cleaned_data["validated_by"] = None
        elif not self.initial["validated"] and cleaned_data.get("validated"):
            cleaned_data["validated_at"] = timezone.now()
            cleaned_data["validated_by"] = self.validator
        del cleaned_data["validated"]
        return cleaned_data


ControlTimesheetFormSet = formset_factory(ControlTimesheetForm, max_num=0, extra=0)
TimesheetFormSet = formset_factory(TimesheetForm, max_num=0, extra=0)


class MonthlyCantonalValidationForm(forms.ModelForm):
    validated = forms.BooleanField(
        label=_("Valider"), required=False, widget=CheckboxInput()
    )

    class Meta:
        model = MonthlyCantonalValidation
        fields = ["validated"]

    def __init__(self, validator, urls, timesheets_statuses, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validator = validator
        self.timesheets_are_validated = timesheets_statuses[self.instance.canton]
        # Add the urls as checkboxes, but first; so:
        # Store away the fields
        fields = self.fields.copy()
        # Empty it
        self.fields = {}
        # Add our fake validations status
        self.fields["timesheets_checked"] = forms.BooleanField(
            label=_("Contrôle des heures effectué"),
            required=False,
            widget=CheckboxInput,
            help_text=mark_safe(
                '<a href="{url}" target="_blank">{text}</a>'.format(
                    url=reverse_lazy(
                        "salary:timesheets-overview",
                        kwargs={"year": self.instance.date.year},
                    ),
                    text=_("Contrôle des heures"),
                )
            ),
            initial=self.timesheets_are_validated,
            disabled=True,
        )
        # urls are MonthlyCantonalValidationUrl
        for url in urls:
            label = url.label if url.label else url.url
            self.fields[f"url_{url.pk}"] = forms.BooleanField(
                label=url.name,
                required=False,
                widget=CheckboxInput,
                help_text=mark_safe(f'<a href="{url.url}" target="_blank">{label}</a>'),
                initial=self.instance.validated_urls.filter(pk=url.pk).exists(),
            )
        # Refill it
        for k, v in fields.items():
            self.fields[k] = v
        if self.initial.get("validated"):
            for key in self.fields:
                self.fields[key].widget.attrs["readonly"] = True
                self.fields[key].disabled = True

    def clean(self):
        cleaned_data = super().clean()
        if self.initial["validated"] and not cleaned_data.get("validated"):
            cleaned_data["validated_at"] = None
            cleaned_data["validated_by"] = None
        elif not self.initial["validated"] and cleaned_data.get("validated"):
            # Verify that all URLs are ticked to allow validation
            all_urls_ticked = True
            for k, v in self.fields.items():
                if k.startswith("url_"):
                    if not cleaned_data.get(k):
                        self.add_error(
                            k,
                            ValidationError(
                                _("Toutes les URLs doivent être vérifiées")
                            ),
                        )
                        all_urls_ticked = False
            if not self.timesheets_are_validated:
                self.add_error(
                    "timesheets_checked",
                    ValidationError(_("Les heures doivent être vérifiées")),
                )
                all_urls_ticked = False

            if all_urls_ticked:
                cleaned_data["validated_at"] = timezone.now()
                cleaned_data["validated_by"] = self.validator
        del cleaned_data["validated"]
        return cleaned_data

    def save(self, commit=True):
        for val in ["validated_at", "validated_by"]:
            try:
                setattr(self.instance, val, self.cleaned_data[val])
            except KeyError:
                pass
        for k, v in self.fields.items():
            if k.startswith("url_"):
                try:
                    mcvu = MonthlyCantonalValidationUrl.objects.get(
                        pk=int(k.split("_")[1])
                    )
                    if self.cleaned_data[k]:
                        self.instance.validated_urls.add(mcvu)
                    else:
                        self.instance.validated_urls.remove(mcvu)
                except (
                    ValueError
                ):  # Conversion of k.split("_")[1] to int; it's not an url
                    pass
        return super().save(commit)



# Generated by Django 2.2.16 on 2020-10-29 10:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("salary", "0012_add_schwyz"),
    ]

    operations = [
        migrations.AddField(
            model_name="timesheet",
            name="ignore",
            field=models.BooleanField(default=False, verbose_name="Ignorer"),
        ),
    ]



from django import template

from ..timesheets_overview import TimesheetStatus

register = template.Library()


@register.filter
def timesheet_status_css_class(timesheet_status):
    """
    Given a `TimesheetStatus`, return a proper CSS class name to represent it.
    """
    if timesheet_status & TimesheetStatus.TIMESHEET_MISSING:
        return "danger"
    elif timesheet_status & TimesheetStatus.TIMESHEET_NOT_VALIDATED:
        return "warning"
    elif timesheet_status & TimesheetStatus.TIMESHEET_VALIDATED:
        return "success"

    return ""



# Generated by Django 1.11.5 on 2017-09-07 10:59
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("challenge", "0048_auto_20170907_1239"),
    ]

    operations = [
        migrations.CreateModel(
            name="HelperSeasonWorkWish",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                (
                    "amount",
                    models.PositiveSmallIntegerField(
                        default=0, verbose_name="Quantit\xe9 d'engagements souhait\xe9s"
                    ),
                ),
                (
                    "helper",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="work_wishes",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Moniteur",
                    ),
                ),
                (
                    "season",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="work_wishes",
                        to="challenge.Season",
                        verbose_name="Saison",
                    ),
                ),
            ],
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date

from django.utils.dates import MONTHS

import factory
from factory import Faker, SubFactory, fuzzy
from factory.django import DjangoModelFactory

from apps.common import DV_SEASON_STATE_OPEN, DV_STATES
from apps.orga.tests.factories import OrganizationFactory
from apps.user.tests.factories import UserFactory

from .. import CHOSEN_AS_ACTOR, CHOSEN_AS_HELPER, CHOSEN_AS_LEADER
from ..models import (
    AnnualStateSetting,
    HelperSessionAvailability,
    Invoice,
    InvoiceLine,
    Qualification,
    QualificationActivity,
    Season,
    Session,
)
from ..models.qualification import CATEGORY_CHOICES
from ..models.registration import Registration


class SeasonFactory(DjangoModelFactory):
    class Meta:
        model = Season

    year = fuzzy.FuzzyInteger(1999, 2050)
    month_start = fuzzy.FuzzyChoice([k for k, v in MONTHS.items()])
    n_months = fuzzy.FuzzyInteger(1, 15)
    # Juste un canton
    cantons = [fuzzy.FuzzyChoice(DV_STATES).fuzz()]
    leader = factory.SubFactory(UserFactory)
    state = DV_SEASON_STATE_OPEN


class SessionFactory(DjangoModelFactory):
    class Meta:
        model = Session

    orga = factory.SubFactory(OrganizationFactory)
    day = fuzzy.FuzzyDate(date(1999, 1, 1), date(2050, 1, 1))


class QualificationFactory(DjangoModelFactory):
    class Meta:
        model = Qualification

    name = Faker("name")
    class_teacher_fullname = Faker("name")
    n_bikes = fuzzy.FuzzyInteger(1, 10)
    n_participants = fuzzy.FuzzyInteger(10, 20)

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        """
        If actor and session is provided, make sure an actor availability object
        exists for this session.
        """
        if "actor" in kwargs and "session" in kwargs:
            HelperSessionAvailability.objects.get_or_create(
                session=kwargs["session"],
                helper=kwargs["actor"],
                availability="y",
                chosen_as=CHOSEN_AS_ACTOR,
            )

        if "leader" in kwargs and "session" in kwargs:
            HelperSessionAvailability.objects.get_or_create(
                session=kwargs["session"],
                helper=kwargs["leader"],
                availability="y",
                chosen_as=CHOSEN_AS_LEADER,
            )
        return super()._create(model_class, *args, **kwargs)

    @factory.post_generation
    def helpers(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for helper in extracted:
                HelperSessionAvailability.objects.get_or_create(
                    session=self.session,
                    helper=helper,
                    availability="y",
                    chosen_as=CHOSEN_AS_HELPER,
                )
                self.helpers.add(helper)


class AnnualStateSettingFactory(DjangoModelFactory):
    class Meta:
        model = AnnualStateSetting

    year = fuzzy.FuzzyInteger(1999, 2050)
    canton = fuzzy.FuzzyChoice(DV_STATES)


class InvoiceFactory(DjangoModelFactory):
    class Meta:
        model = Invoice

    season = SubFactory(SeasonFactory)
    organization = SubFactory(OrganizationFactory)
    ref = fuzzy.FuzzyText(length=20)
    status = fuzzy.FuzzyChoice([s[0] for s in Invoice.STATUS_CHOICES])


class InvoiceLineFactory(DjangoModelFactory):
    class Meta:
        model = InvoiceLine

    session = SubFactory(SessionFactory)
    invoice = SubFactory(InvoiceFactory)
    nb_bikes = fuzzy.FuzzyInteger(0, 20)
    nb_participants = fuzzy.FuzzyInteger(0, 20)
    cost_bikes = fuzzy.FuzzyDecimal(0, 400)
    cost_participants = fuzzy.FuzzyDecimal(0, 400)


class QualificationActivityFactory(DjangoModelFactory):
    class Meta:
        model = QualificationActivity

    name = Faker("name")
    category = fuzzy.FuzzyChoice([c[0] for c in CATEGORY_CHOICES])


class RegistrationFactory(DjangoModelFactory):
    class Meta:
        model = Registration



# Generated by Django 1.11.4 on 2017-08-29 07:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("orga", "0004_auto_20170817_1424"),
    ]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="abbr",
            field=models.CharField(
                blank=True, max_length=16, verbose_name="Abbréviation"
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="status",
            field=models.PositiveSmallIntegerField(
                choices=[(0, "---------"), (10, "Actif"), (30, "Inactif")],
                default=10,
                verbose_name="Statut",
            ),
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-18 09:12

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("salary", "0007_auto_20200409_0249"),
    ]

    operations = [
        migrations.CreateModel(
            name="MonthlyCantonalValidation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateField(verbose_name="Date (1er du mois)")),
                (
                    "canton",
                    models.CharField(
                        choices=[
                            ("AG", "Aargau"),
                            ("AR", "Appenzell Ausserrhoden"),
                            ("BS", "Basel-Stadt"),
                            ("BL", "Basel-Land"),
                            ("BE", "Berne"),
                            ("FR", "Fribourg"),
                            ("GE", "Geneva"),
                            ("GR", "Graubuenden"),
                            ("JU", "Jura"),
                            ("LU", "Lucerne"),
                            ("NE", "Neuchatel"),
                            ("SO", "Solothurn"),
                            ("SG", "St. Gallen"),
                            ("VS", "Valais"),
                            ("VD", "Vaud"),
                            ("ZH", "Zurich"),
                            ("WS", "Haut-Valais"),
                        ],
                        max_length=2,
                        unique_for_month="date",
                        unique_for_year="date",
                        verbose_name="Canton",
                    ),
                ),
                ("validated_at", models.DateTimeField(blank=True, null=True)),
                (
                    "validated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Validation mensuelle cantonale",
                "verbose_name_plural": "Validations mensuelles cantonales",
                "ordering": ["date", "canton"],
                "unique_together": {("canton", "date")},
            },
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0020_auto_20160926_1655"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="userprofile",
            name="office_member",
        ),
    ]



# Generated by Django 1.11.4 on 2017-09-05 10:30
from __future__ import unicode_literals

from django.contrib.postgres.operations import UnaccentExtension
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0005_VS-OW_to_WS"),
    ]

    operations = [UnaccentExtension()]



from django.db import migrations, models


def ws_to_vs(apps, schema_editor):
    MonthlyCantonalValidation = apps.get_model("salary", "MonthlyCantonalValidation")
    for mcv in MonthlyCantonalValidation.objects.filter(canton="WS"):
        # As of 2020-11-09, there are no _validated_ WS mcvs.
        # Still check if there's a VS one (there will), and delete.
        if MonthlyCantonalValidation.objects.filter(
            canton="VS", date=mcv.date
        ).exists():
            mcv.delete()
        else:
            mcv.canton = "VS"
            mcv.save()


class Migration(migrations.Migration):
    dependencies = [
        ("salary", "0013_add_ignore_to_timesheets"),
    ]

    operations = [
        migrations.RunPython(ws_to_vs, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="monthlycantonalvalidation",
            name="canton",
            field=models.CharField(
                choices=[
                    ("AG", "Aargau"),
                    ("AR", "Appenzell Ausserrhoden"),
                    ("BS", "Basel-Stadt"),
                    ("BL", "Basel-Land"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("GR", "Graubuenden"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SZ", "Schwyz"),
                    ("SO", "Solothurn"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                ],
                max_length=2,
                unique_for_month="date",
                unique_for_year="date",
                verbose_name="Canton",
            ),
        ),
    ]



# Generated by Django 2.2.17 on 2021-01-11 09:54

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0077_auto_20201214_1704"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalqualification",
            name="n_bikes",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de vélos",
            ),
        ),
        migrations.AlterField(
            model_name="historicalqualification",
            name="n_helmets",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de casques",
            ),
        ),
        migrations.AlterField(
            model_name="historicalqualification",
            name="n_participants",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de participants",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="n_bikes",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de vélos",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="n_helmets",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de casques",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="n_participants",
            field=models.PositiveSmallIntegerField(
                blank=True,
                null=True,
                validators=[django.core.validators.MaxValueValidator(30)],
                verbose_name="Nombre de participants",
            ),
        ),
    ]



from .timesheets import (
    CleanupOrphanedTimesheets,
    ExportMonthlyControl,
    ExportMonthlyTimesheets,
    RedirectUserMonthlyTimesheets,
    SendTimesheetsReminder,
    UserMonthlyTimesheets,
    YearlyTimesheets,
)
from .validations import ValidationsMonthView, ValidationsYearView, ValidationUpdate

__all__ = [
    "CleanupOrphanedTimesheets",
    "ExportMonthlyControl",
    "ExportMonthlyTimesheets",
    "RedirectUserMonthlyTimesheets",
    "SendTimesheetsReminder",
    "UserMonthlyTimesheets",
    "ValidationUpdate",
    "ValidationsMonthView",
    "ValidationsYearView",
    "YearlyTimesheets",
]



from datetime import time

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from apps.challenge.models import Session
from apps.orga.models import Organization

User = get_user_model()

REGISTRATION_MORNING = "08:30:00"
REGISTRATION_AFTERNOON = "13:30:00"

REGISTRATION_DAY_TIMES = (
    (REGISTRATION_MORNING, _("Matin")),
    (REGISTRATION_AFTERNOON, _("Après-midi")),
)


class Registration(models.Model):
    coordinator = models.ForeignKey(User, on_delete=models.CASCADE)
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    date = models.DateField(_("Date"))
    day_time = models.CharField(
        verbose_name=_("Moment de la journée"),
        choices=REGISTRATION_DAY_TIMES,
        default=REGISTRATION_MORNING,
        max_length=8,
    )
    classes_amount = models.IntegerField(
        verbose_name=_("Nb de classes (1 à 6)"),
        default=1,
        validators=[MinValueValidator(1), MaxValueValidator(6)],
    )

    is_archived = models.BooleanField(verbose_name=_("Archivée"), default=False)

    def clean(self):
        if not self.coordinator.managed_organizations.filter(
            pk=self.organization.pk
        ).exists():
            raise ValidationError(
                {
                    "organization": _(
                        "L'établissement choisi n'est pas géré par cet utilisateur·rice."
                    )
                }
            )

        if Session.sessions_in_same_half_day(
            self.organization, self.date, self.day_time_as_time
        ).exists():
            raise ValidationError(
                {
                    "date": _(
                        "Il y a déjà une session inscrite pour cet établissement à cette date."
                    )
                }
            )

    def archive(self):
        self.is_archived = True
        self.save()

    @property
    def day_time_as_time(self):
        as_array = self.day_time.split(":")
        return time(int(as_array[0]), int(as_array[1]))



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.core.exceptions import PermissionDenied, ValidationError
from django.forms import Form as DjangoEmptyForm
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormView

from rolepermissions.checkers import has_role
from rolepermissions.mixins import HasPermissionsMixin

from defivelo.roles import has_permission, user_cantons

from ..forms import UserAssignRoleForm
from .mixins import ProfileMixin


class UserCredentials(ProfileMixin, FormView):
    template_name = "auth/user_send_credentials.html"
    success_message = _("Données de connexion expédiées.")
    form_class = DjangoEmptyForm
    initial_send = True

    def dispatch(self, request, *args, **kwargs):
        self.managed_cantons = user_cantons(self.request.user)
        self.user_cantons_intersection = [
            orga.address_canton
            for orga in self.get_object().managed_organizations.all()
            if orga.address_canton in self.managed_cantons
        ]
        if (
            self.get_object().profile.affiliation_canton in self.managed_cantons
            or self.user_cantons_intersection
            # Useful when the edited user has no canton (affiliation_canton == '')
            or has_permission(self.request.user, "cantons_all")
        ):
            return super().dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied

    def get_context_data(self, **kwargs):
        context = super(UserCredentials, self).get_context_data(**kwargs)

        user_cantons_intersection = self.user_cantons_intersection + [
            self.get_object().profile.affiliation_canton
        ]
        context["userprofile"] = self.get_object()
        context["userprofilecanton"] = user_cantons_intersection
        context["initial_send"] = self.initial_send
        return context

    def form_valid(self, form):
        context = self.get_context_data()
        context["fromuser"] = self.request.user
        user = self.get_object()
        user.profile.send_credentials(context, force=(not self.initial_send))
        return super(UserCredentials, self).form_valid(form)


class SendUserCredentials(HasPermissionsMixin, UserCredentials):
    required_permission = "user_can_send_credentials"
    cantons = False

    def dispatch(self, request, *args, **kwargs):
        # Forbid view if can already login
        if not self.get_object().profile.can_login:
            return super(SendUserCredentials, self).dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied


class ResendUserCredentials(HasPermissionsMixin, UserCredentials):
    required_permission = "user_can_resend_credentials"
    success_message = _("Données de connexion ré-expédiées.")
    initial_send = False
    cantons = False

    def dispatch(self, request, *args, **kwargs):
        # Forbid view if initial login hasn't been sent
        if self.get_object().profile.can_login and not self.initial_send:
            return super(ResendUserCredentials, self).dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied


class UserAssignRole(ProfileMixin, FormView):
    template_name = "roles/assign.html"
    success_message = _("Rôle assigné à l’utilisa·teur·trice")
    form_class = UserAssignRoleForm
    cantons = False

    def dispatch(self, request, *args, **kwargs):
        # Forbid if self
        user = self.get_object()

        if user != self.request.user and user.profile.can_login:
            if has_permission(self.request.user, "user_set_role") or has_permission(
                self.request.user, "assign_only_coordinator_role"
            ):
                return super(UserAssignRole, self).dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied

    def get_context_data(self, **kwargs):
        context = super(UserAssignRole, self).get_context_data(**kwargs)
        context["userprofile"] = self.get_object()
        context["user_has_a_role"] = self.get_object().groups.exists()
        context["profile_is_coordinator"] = has_role(self.get_object(), "coordinator")
        context["requester_is_state_manager"] = has_role(
            self.request.user, "state_manager"
        )
        return context

    def get_form_kwargs(self):
        form_kwargs = super().get_form_kwargs()
        form_kwargs.update({"user": self.get_object()})
        return form_kwargs

    def form_valid(self, form):
        if (
            has_role(self.get_object(), "coordinator")
            and has_permission(self.request.user, "assign_only_coordinator_role")
            and not has_permission(self.request.user, "user_set_role")
        ):
            if not form["role"].data == "coordinator":
                raise ValidationError(
                    _("vous ne pouvez pas annuler l'attribution du rôle")
                )

        form.save()
        return super(UserAssignRole, self).form_valid(form)



from django.contrib.auth import get_user_model
from django.db import migrations
from rolepermissions.checkers import has_role
from rolepermissions.roles import assign_role, clear_roles


def reset_all_roles(apps, schema_editor):
    User = get_user_model()
    for user in User.objects.all():
        for role in ["collaborator", "state_manager", "power_user", "coordinator"]:
            if has_role(user, role):
                clear_roles(user)
                assign_role(user, role)
                continue


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0065_role_reset_for_orga_edit_address"),
    ]

    operations = [
        migrations.RunPython(reset_all_roles, migrations.RunPython.noop),
    ]



# Generated by Django 1.11.4 on 2017-08-24 09:09
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0036_auto_20170818_1633"),
    ]

    operations = [
        migrations.AlterField(
            model_name="qualification",
            name="activity_A",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_A",
                to="challenge.QualificationActivity",
                verbose_name="Vélo urbain",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="activity_B",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_B",
                to="challenge.QualificationActivity",
                verbose_name="Mécanique",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="activity_C",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_C",
                to="challenge.QualificationActivity",
                verbose_name="Rencontre",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="actor",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_actor",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Intervenant",
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="leader",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_mon2",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Moniteur 2",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="cantons",
            field=models.CharField(
                choices=[
                    ("AP", "Appenzell"),
                    ("BE", "Berne"),
                    ("BS", "Basel-Stadt"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("VS-OW", "Haut-Valais"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                ],
                max_length=41,
                verbose_name="Cantons",
            ),
        ),
        migrations.AlterField(
            model_name="session",
            name="superleader",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="sess_monplus",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Moniteur + / Photographe",
            ),
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-12 14:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0064_alter_invoiceline_ordering_again"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalsession",
            name="visible",
            field=models.BooleanField(
                default=False, verbose_name="Visible pour les moniteurs"
            ),
        ),
        migrations.AddField(
            model_name="session",
            name="visible",
            field=models.BooleanField(
                default=False, verbose_name="Visible pour les moniteurs"
            ),
        ),
    ]



# Generated by Django 4.2.6 on 2025-09-16 06:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('challenge', '0082_add_states_ow_nw'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalqualification',
            name='n_helpers',
            field=models.IntegerField(choices=[(1, '1'), (2, '2'), (3, '3')], default=3, verbose_name='Nombre de moniteurs'),
        ),
        migrations.AddField(
            model_name='qualification',
            name='n_helpers',
            field=models.IntegerField(choices=[(1, '1'), (2, '2'), (3, '3')], default=3, verbose_name='Nombre de moniteurs'),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from rolepermissions.mixins import HasPermissionsMixin

from apps.article.models import Article
from defivelo.views import MenuView

from .forms import ArticleForm


class ArticleMixin(HasPermissionsMixin, SuccessMessageMixin, MenuView):
    required_permission = "home_article_cud"
    model = Article
    context_object_name = "article"
    success_url = reverse_lazy("home")


class ArticleInputMixin(ArticleMixin):
    form_class = ArticleForm


class ArticleCreateView(ArticleInputMixin, CreateView):
    success_message = _("Article créé")


class ArticleUpdateView(ArticleInputMixin, UpdateView):
    success_message = _("Article mis à jour")


class ArticleDeleteView(ArticleMixin, DeleteView):
    pass



# Generated by Django 1.11.4 on 2017-09-04 15:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0044_season_state"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="season",
            options={
                "ordering": ["year", "-season", "cantons"],
                "verbose_name": "Saison",
                "verbose_name_plural": "Saisons",
            },
        ),
        migrations.AddField(
            model_name="session",
            name="bikes_concept",
            field=models.CharField(
                blank=True, max_length=512, verbose_name="Logistique v\xe9los"
            ),
        ),
        migrations.AddField(
            model_name="session",
            name="bikes_phone",
            field=models.CharField(
                blank=True, max_length=13, verbose_name="N\xb0 de contact v\xe9los"
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from urllib.parse import quote

from django.conf import settings
from django.contrib.messages.views import SuccessMessageMixin
from django.contrib.sites.models import Site
from django.core.exceptions import FieldError, PermissionDenied
from django.template.defaultfilters import date, time
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.html import urlencode
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.views.generic.dates import WeekArchiveView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from tablib import Dataset

from apps.common.views import ExportMixin
from defivelo.roles import has_permission, user_cantons
from defivelo.templatetags.dv_filters import lettercounter
from defivelo.views import MenuView

from ..forms import QualificationFormQuick, SessionForm
from ..forms.session import SessionDeleteForm
from ..models import Session
from ..models.qualification import (
    CATEGORY_CHOICE_A,
    CATEGORY_CHOICE_B,
    CATEGORY_CHOICE_C,
)
from .mixins import CantonSeasonFormMixin

EXPORT_NAMETEL = gettext("{name} - {tel}")


class SessionMixin(CantonSeasonFormMixin, MenuView):
    required_permission = "challenge_session_crud"
    model = Session
    context_object_name = "session"
    form_class = SessionForm
    raise_without_cantons = False

    def get_queryset(self):
        qs = super().get_queryset()
        try:
            return qs.filter(orga__address_canton__in=self.season_object.cantons)
        except FieldError:
            # For the cases qs is Qualification, not Session
            return qs

    def get_success_url(self):
        return reverse_lazy(
            "session-detail",
            kwargs={"seasonpk": self.season_object.pk, "pk": self.object.pk},
        )

    def get_context_data(self, **kwargs):
        context = super(SessionMixin, self).get_context_data(**kwargs)
        # Add our menu_category context
        context["menu_category"] = "season"
        context["season"] = self.season_object
        try:
            mysession = self.get_object()
        except AttributeError:
            return context

        session_pages = {
            "session_current": None,
            "session_next": None,
        }

        try:
            sessions = self.season.sessions_by_orga.filter(orga=mysession.orga)
        except AttributeError:
            sessions = []

        # Iterate through all of them, index is 'next'
        for session in sessions:
            session_pages["session_current"] = session_pages["session_next"]
            session_pages["session_next"] = session
            # On arrive
            if session_pages["session_next"].pk == mysession.pk:
                context["session_previous"] = session_pages["session_current"]
            # On y est
            if (
                session_pages["session_current"]
                and session_pages["session_current"].pk == mysession.pk
            ):
                context["session_next"] = session_pages["session_next"]
                break
        return context


class SessionsListView(SessionMixin, WeekArchiveView):
    date_field = "day"
    context_object_name = "sessions"
    allow_empty = True
    allow_future = True
    week_format = "%W"
    ordering = ["day", "begin", "duration"]

    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for access to list
        """
        allowed = False

        if has_permission(request.user, self.required_permission):
            # StateManagers
            # Check that the intersection isn't empty
            usercantons = user_cantons(request.user)
            if usercantons:
                if list(set(usercantons).intersection(set(self.season_object.cantons))):
                    # StateManager cantons
                    allowed = True
                else:
                    # If the user is marked as state manager for that season
                    if self.season_object.leader == self.request.user:
                        allowed = True
                    # Verify that this state manager can access that canton as mobile
                    elif list(
                        set(
                            [request.user.profile.affiliation_canton]
                            + request.user.profile.activity_cantons
                        ).intersection(set(self.season_object.cantons))
                    ):
                        allowed = True
        elif self.season_object.unprivileged_user_can_see(request.user):
            allowed = True

        if allowed:
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied


class SessionDetailView(SessionMixin, DetailView):
    # Allow season fetch even for non-state managers
    allow_season_fetch = True

    @property
    def print_filename(self):
        return _("Session") + "_" + force_str(self.object.get_export_filename())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        try:
            mysession = self.get_object()
        except AttributeError:
            return context
        context["quali_form_quick"] = QualificationFormQuick(
            {
                "session": mysession,
                "name": _("Classe %s") % lettercounter(mysession.n_qualifications + 1),
            }
        )
        context["statemanager_can_access"] = (
            self.season_object.manager_can_crud
            and has_permission(self.request.user, "challenge_session_crud")
        )
        context["coordinator_can_access"] = (
            self.season_object.coordinator_can_update
            and mysession.orga.coordinator == self.request.user
        )
        context["user_is_coordinator"] = mysession.orga.coordinator == self.request.user
        # Build a meaningful mailto: link towards all session available emails, with meaningful subject and body.
        session_helpers = [
            f"{s.helper.get_full_name()} <{s.helper.email}>"
            for s in mysession.chosen_staff.order_by(
                "helper__first_name", "helper__last_name"
            )
        ]
        context["session_mailtoall"] = None
        if session_helpers:
            emailparts = {
                "subject": (
                    settings.EMAIL_SUBJECT_PREFIX
                    + _("Qualifs {session}").format(session=mysession)
                ),
                "body": render_to_string(
                    "challenge/session_email_to_helpers.txt",
                    {
                        "current_site": Site.objects.get_current(),
                        "session": mysession,
                        "session_url": self.request.build_absolute_uri(
                            reverse(
                                "session-detail",
                                kwargs={
                                    "seasonpk": mysession.season.pk,
                                    "pk": mysession.pk,
                                },
                            )
                        ),
                    },
                ),
            }
            context["session_mailtoall"] = "mailto:{emaillist}?{options}".format(
                emaillist=", ".join(session_helpers),
                options=urlencode(emailparts, quote_via=quote),
            )

        context["pdf_file_name"] = _("DV-{print_filename}_{YMD_date}").format(
            print_filename=self.print_filename,
            YMD_date=timezone.now().strftime("%Y%m%d"),
        )

        return context

    def get_queryset(self):
        return (
            super(SessionDetailView, self)
            .get_queryset()
            .prefetch_related(
                "orga",
                "qualifications",
                "qualifications__leader",
                "qualifications__leader__profile",
                "qualifications__helpers",
                "qualifications__helpers__profile",
                "qualifications__actor",
                "qualifications__actor__profile",
                "qualifications__activity_A",
                "qualifications__activity_B",
                "qualifications__activity_C",
            )
        )

    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for access to view
        """
        allowed = False

        if has_permission(request.user, self.required_permission):
            allowed = True
        else:
            # Read-only view when session is visible
            if self.season and self.season.unprivileged_user_can_see(request.user):
                session = self.get_object()
                # Always visible for coordinators
                if session.visible or session.orga.coordinator == request.user:
                    allowed = True

        if allowed:
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied


class SessionCloneView(SessionMixin, SuccessMessageMixin, UpdateView):
    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for cloning
        """
        if (
            has_permission(request.user, self.required_permission)
            and self.season
            and self.season.manager_can_crud
        ):
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def get_form_kwargs(self):
        args = super().get_form_kwargs()
        if "instance" in args:
            self._clear_clone_fields(args.get("instance"))

        return args

    def post(self, request, *args, **kwargs):
        raise PermissionDenied('POST not allowed, use "create" url instead')

    def _clear_clone_fields(self, obj: Session):
        obj.pk = None
        if self.request.method.lower() == "get":
            obj.day = None
            obj.begin = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["action"] = reverse_lazy(
            "session-create", kwargs={"seasonpk": self.kwargs.get("seasonpk")}
        )
        return context


class SessionUpdateView(SessionMixin, SuccessMessageMixin, UpdateView):
    success_message = _("Session mise à jour")

    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for access to updateview
        """
        allowed = False

        if (
            has_permission(request.user, self.required_permission)
            and self.season
            and self.season.manager_can_crud
        ):
            # StateManager, and it's the right moment
            allowed = True
        elif (
            self.get_object().orga.coordinator == request.user
            and self.season_object.coordinator_can_update
        ):
            # Coordinator, and it's the right moment
            allowed = True

        if allowed:
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        # Coordinator without StateManager rights
        kwargs["is_for_coordinator"] = (
            not has_permission(self.request.user, "challenge_session_crud")
            and self.request.user == self.get_object().orga.coordinator
        )
        return kwargs


class SessionCreateView(SessionMixin, SuccessMessageMixin, CreateView):
    success_message = _("Session créée")

    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for access to view
        """
        if (
            has_permission(request.user, self.required_permission)
            and self.season
            and self.season.manager_can_crud
        ):
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied


class SessionDeleteView(SessionMixin, SuccessMessageMixin, DeleteView):
    success_message = _("Session supprimée")
    form_class = SessionDeleteForm

    def dispatch(self, request, *args, **kwargs):
        """
        Check allowances for access to view
        """
        if (
            has_permission(request.user, self.required_permission)
            and self.season
            and self.season.manager_can_crud
        ):
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def form_valid(self, form):
        self.get_object().get_related_timesheets().delete()
        return super().form_valid(form)

    def get_success_url(self):
        return reverse_lazy("season-detail", kwargs={"pk": self.kwargs["seasonpk"]})


class SessionStaffChoiceView(SessionDetailView):
    template_name = "challenge/session_availability.html"


class SessionExportView(ExportMixin, SessionMixin, DetailView):
    # Allow season fetch even for non-state managers
    allow_season_fetch = True

    @property
    def export_filename(self):
        return _("Session") + "_" + force_str(self.object.get_export_filename())

    def get_dataset(self):
        dataset = Dataset()
        # Prépare le fichier
        dataset.append_col(
            [
                gettext("Date"),
                gettext("Canton"),
                gettext("Établissement"),
                gettext("Emplacement"),
                gettext("Heures"),
                gettext("Nombre de qualifs"),
                # Logistique
                gettext("Moniteur·trice + / Photographe"),
                gettext("Mauvais temps"),
                gettext("Logistique vélos"),
                gettext("N° de contact vélos"),
                gettext("Pommes"),
                gettext("Total vélos"),
                gettext("Total casques"),
                gettext("Remarques"),
                # Qualif
                gettext("Classe"),
                gettext("Enseignant·e"),
                gettext("Moniteur·trice 2"),
                gettext("Moniteur·trice 1"),
                gettext("Moniteur·trice 1"),
                gettext("Nombre d’élèves"),
                gettext("Nombre de vélos"),
                gettext("Nombre de casques"),
                CATEGORY_CHOICE_A,
                CATEGORY_CHOICE_B,
                CATEGORY_CHOICE_C,
                gettext("Intervenant·e"),
                gettext("Remarques"),
            ]
        )
        session = self.object
        session_place = session.place
        if not session_place:
            session_place = (
                session.address_city
                if session.address_city
                else session.orga.address_city
            )
        col = [
            date(session.day),
            session.orga.address_canton,
            session.orga.name,
            session_place,
            "%s - %s" % (time(session.begin), time(session.end)),
            session.n_qualifications,
            EXPORT_NAMETEL.format(
                name=session.superleader.get_full_name(),
                tel=session.superleader.profile.natel,
            )
            if session.superleader
            else "",
            str(session.fallback),
            session.bikes_concept,
            session.bikes_phone,
            session.apples,
            session.n_bikes,
            session.n_helmets,
            session.comments,
        ]
        if session.n_qualifications:
            for quali in session.qualifications.all():
                if not len(col):
                    col = [""] * 14
                col.append(quali.name)
                col.append(
                    EXPORT_NAMETEL.format(
                        name=quali.class_teacher_fullname, tel=quali.class_teacher_natel
                    )
                    if quali.class_teacher_fullname
                    else ""
                )
                col.append(
                    EXPORT_NAMETEL.format(
                        name=quali.leader.get_full_name(),
                        tel=quali.leader.profile.natel,
                    )
                    if quali.leader
                    else ""
                )
                for i in range(2):
                    try:
                        helper = quali.helpers.all()[i]
                        col.append(
                            EXPORT_NAMETEL.format(
                                name=helper.get_full_name(), tel=helper.profile.natel
                            )
                            if helper
                            else ""
                        )
                    except IndexError:
                        col.append("")
                col.append(quali.n_participants)
                col.append(quali.n_bikes)
                col.append(quali.n_helmets)
                col.append(str(quali.activity_A) if quali.activity_A else "")
                col.append(str(quali.activity_B) if quali.activity_B else "")
                col.append(str(quali.activity_C) if quali.activity_C else "")
                col.append(
                    EXPORT_NAMETEL.format(
                        name=quali.actor.get_full_name(), tel=quali.actor.profile.natel
                    )
                    if quali.actor
                    else ""
                )
                col.append(quali.comments)
                dataset.append_col(col)
                col = []
        else:
            col += [""] * 13
            dataset.append_col(col)
        return dataset



# Generated by Django 2.2.12 on 2020-05-20 07:05

import localflavor.generic.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0057_orga_detail_all_perm"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="iban",
            field=localflavor.generic.models.IBANField(
                blank=True,
                include_countries=(
                    "AT",
                    "BE",
                    "BG",
                    "CH",
                    "CY",
                    "CZ",
                    "DE",
                    "DK",
                    "EE",
                    "ES",
                    "FI",
                    "FR",
                    "GB",
                    "GI",
                    "GR",
                    "HR",
                    "HU",
                    "IE",
                    "IS",
                    "IT",
                    "LI",
                    "LT",
                    "LU",
                    "LV",
                    "MC",
                    "MT",
                    "NL",
                    "NO",
                    "PL",
                    "PT",
                    "RO",
                    "SE",
                    "SI",
                    "SK",
                    "SM",
                ),
                max_length=34,
                use_nordea_extensions=False,
                verbose_name="Coordonnées bancaires (IBAN)",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="social_security",
            field=models.CharField(blank=True, max_length=16, verbose_name="N° AVS"),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="tax_jurisdiction",
            field=models.CharField(
                blank=True, max_length=511, verbose_name="Lieu d'imposition"
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="work_permit",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="Permis de travail"
            ),
        ),
    ]



# Generated by Django 1.11.4 on 2017-08-18 14:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0032_set_deletion_roles"),
    ]

    operations = [
        migrations.AlterField(
            model_name="usermanagedstate",
            name="canton",
            field=models.CharField(
                choices=[
                    ("BS", "Basel-Stadt"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                    ("VS-OW", "Haut-Valais"),
                    ("AP", "Appenzell"),
                ],
                max_length=5,
                verbose_name="Canton",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="activity_cantons",
            field=models.CharField(
                blank=True,
                choices=[
                    ("BS", "Basel-Stadt"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                    ("VS-OW", "Haut-Valais"),
                    ("AP", "Appenzell"),
                ],
                max_length=41,
                verbose_name="Défi Vélo mobile",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="affiliation_canton",
            field=models.CharField(
                blank=True,
                choices=[
                    ("", "---------"),
                    ("BS", "Basel-Stadt"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                    ("VS-OW", "Haut-Valais"),
                    ("AP", "Appenzell"),
                ],
                max_length=5,
                verbose_name="Canton d'affiliation",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="bagstatus",
            field=models.PositiveSmallIntegerField(
                choices=[(0, "---"), (10, "En prêt"), (20, "Payé"), (20, "Offert")],
                default=0,
                verbose_name="Sac Défi Vélo",
            ),
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from parler.models import TranslatableModel, TranslatedFields

from apps.common import DV_STATE_CHOICES


class MonthlyCantonalValidationUrl(TranslatableModel):
    translations = TranslatedFields(
        name=models.CharField(_("Nom"), max_length=255),
        url=models.URLField(_("URL")),
        label=models.CharField(_("Label"), max_length=255, blank=True),
    )

    class Meta:
        verbose_name = _("URL pour validation mensuelle")
        verbose_name_plural = _("URLs pour validation mensuelle")

    def __str__(self):
        return self.name


class MonthlyCantonalValidation(models.Model):
    date = models.DateField("Date (1er du mois)")
    canton = models.CharField(
        _("Canton"),
        max_length=2,
        choices=DV_STATE_CHOICES,
        unique_for_month="date",
        unique_for_year="date",
    )
    validated_at = models.DateTimeField(null=True, blank=True)
    validated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="+",
        on_delete=models.SET_NULL,
        null=True,
    )
    validated_urls = models.ManyToManyField(MonthlyCantonalValidationUrl)

    class Meta:
        unique_together = (
            (
                "canton",
                "date",
            ),
        )
        verbose_name = _("Validation mensuelle cantonale")
        verbose_name_plural = _("Validations mensuelles cantonales")
        ordering = ["date", "canton"]

    def __str__(self):
        return _("{year}/{month}: Validation du canton {canton}").format(
            year=self.date.year, month=self.date.month, canton=self.canton
        )

    def save(self, *args, **kwargs):
        """
        Enforce date to get day=1
        """
        self.date = self.date.replace(day=1)
        super().save(*args, **kwargs)

    @property
    def validated(self):
        if self.validated_at is None:
            return False
        if self.validated_by is None:
            return False

        our_urls_pks = self.validated_urls.all().values_list("pk", flat=True)
        all_urls_pks = MonthlyCantonalValidationUrl.objects.all().values_list(
            "pk", flat=True
        )
        for url_pk in all_urls_pks:
            if url_pk not in our_urls_pks:
                return False
        return True



import datetime

from django.urls import reverse

from rolepermissions.roles import assign_role

from apps.orga.tests.factories import OrganizationFactory
from apps.user.tests.factories import UserFactory
from defivelo.tests.utils import CoordinatorAuthClient, StateManagerAuthClient

from ..forms.registration import RegistrationFormSet
from ..models import Qualification, Session
from ..models.registration import Registration
from .factories import RegistrationFactory
from .test_seasons import SeasonTestCaseMixin


class RegistrationTestCase(SeasonTestCaseMixin):
    def setUp(self):
        self.client = CoordinatorAuthClient()
        super().setUp()

        # Put the sessions for our Coordinator's orga
        for s in self.sessions:
            s.orga.coordinator = self.client.user
            s.orga.save()

        # Specify one Session's moment
        self.sessions[0].day = datetime.date(2021, 1, 10)
        self.sessions[0].begin = "09:00:00"
        self.sessions[0].save()

        # Specify Season moment
        self.season.year = 2021
        self.season.month_start = 1
        self.season.n_months = 3
        self.season.save()

    def test_coordinator_sees_registrations_form(self):
        response = self.client.get(reverse("registration-create"))
        assert response.status_code == 200
        assert isinstance(response.context.get("formset", None), RegistrationFormSet)

    def test_coordinator_can_submit_registrations_in_season(self):
        organization = self.client.user.managed_organizations.first()
        response = self.client.post(
            reverse("registration-create"),
            data={
                "organization": organization.pk,
                "form-TOTAL_FORMS": "2",
                "form-INITIAL_FORMS": "3",
                "form-MIN_NUM_FORMS": "1",
                "form-MAX_NUM_FORMS": "1",
                "form-0-date": "20.01.2021",
                "form-0-day_time": "13:30:00",
                "form-0-classes_amount": "1",
                "form-1-date": "21.01.2021",
                "form-1-day_time": "08:30:00",
                "form-1-classes_amount": "2",
                "form-__prefix__-date": "",
                "form-__prefix__-day_time": "08:30:00",
                "form-__prefix__-classes_amount": "1",
            },
        )

        assert response.status_code == 302
        assert response.url == reverse("registration-confirm")

    def test_coordinator_cant_submit_registrations_out_of_season(self):
        organization = self.client.user.managed_organizations.first()
        response = self.client.post(
            reverse("registration-create"),
            data={
                "organization": organization.pk,
                "form-TOTAL_FORMS": "2",
                "form-INITIAL_FORMS": "3",
                "form-MIN_NUM_FORMS": "1",
                "form-MAX_NUM_FORMS": "1",
                "form-0-date": "20.04.2021",
                "form-0-day_time": "13:30:00",
                "form-0-classes_amount": "1",
                "form-1-date": "21.05.2021",
                "form-1-day_time": "08:30:00",
                "form-1-classes_amount": "2",
                "form-__prefix__-date": "",
                "form-__prefix__-day_time": "08:30:00",
                "form-__prefix__-classes_amount": "1",
            },
        )

        assert response.status_code == 200
        assert len(response.context["formset"].errors) == 2

    def test_coordinator_cant_submit_parallel_sessions(self):
        organization = self.client.user.managed_organizations.first()
        response = self.client.post(
            reverse("registration-create"),
            data={
                "organization": organization.pk,
                "form-TOTAL_FORMS": "2",
                "form-INITIAL_FORMS": "3",
                "form-MIN_NUM_FORMS": "1",
                "form-MAX_NUM_FORMS": "1",
                "form-0-date": "20.02.2021",
                "form-0-day_time": "13:30:00",
                "form-0-classes_amount": "1",
                "form-1-date": "20.02.2021",
                "form-1-day_time": "13:30:00",
                "form-1-classes_amount": "2",
                "form-__prefix__-date": "",
                "form-__prefix__-day_time": "08:30:00",
                "form-__prefix__-classes_amount": "1",
            },
        )

        assert response.status_code == 200
        assert "date" in response.context["formset"].errors[1]

    def test_coordinator_cant_submit_sessions_over_existing_others(self):
        organization = self.client.user.managed_organizations.first()
        response = self.client.post(
            reverse("registration-create"),
            data={
                "organization": organization.pk,
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "0",
                "form-MIN_NUM_FORMS": "1",
                "form-MAX_NUM_FORMS": "1",
                "form-0-date": "10.01.2021",
                "form-0-day_time": "08:30:00",
                "form-0-classes_amount": "1",
                "form-__prefix__-date": "",
                "form-__prefix__-day_time": "08:30:00",
                "form-__prefix__-classes_amount": "1",
            },
        )

        assert response.status_code == 200
        assert len(response.context["formset"].errors) == 1

    def test_coordinator_cant_submit_sessions_to_not_owned_organization(self):
        organization = OrganizationFactory()
        response = self.client.post(
            reverse("registration-create"),
            data={
                "organization": organization.pk,
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "0",
                "form-MIN_NUM_FORMS": "1",
                "form-MAX_NUM_FORMS": "1",
                "form-0-date": "20.01.2021",
                "form-0-day_time": "08:30:00",
                "form-0-classes_amount": "1",
                "form-__prefix__-date": "",
                "form-__prefix__-day_time": "08:30:00",
                "form-__prefix__-classes_amount": "1",
            },
        )

        assert response.status_code == 200
        assert "organization" in response.context["form"].errors

    def test_coordinator_must_accept_conditions(self):
        organization = self.client.user.managed_organizations.first()
        session = self.client.session
        session.update(
            {
                "new_registration": {
                    "organization": organization.pk,
                    "lines": [
                        {
                            "date": "2021-01-20",
                            "day_time": "08:30:00",
                            "classes_amount": 1,
                        }
                    ],
                }
            }
        )
        session.save()
        response = self.client.post(reverse("registration-confirm"), data={})
        assert response.status_code == 200
        assert "is_terms_accepted" in response.context["form"].errors

    def test_coordinator_can_confirm_registrations(self):
        organization = self.client.user.managed_organizations.first()
        http_session = self.client.session
        http_session.update(
            {
                "new_registration": {
                    "organization": organization.pk,
                    "lines": [
                        {
                            "date": "2021-01-20",
                            "day_time": "08:30:00",
                            "classes_amount": 1,
                        }
                    ],
                }
            }
        )
        http_session.save()

        response = self.client.post(
            reverse("registration-confirm"), data={"is_terms_accepted": "on"}
        )
        assert response.status_code == 302

        assert (
            Registration.objects.filter(
                date="2021-01-20",
                day_time="08:30:00",
                classes_amount=1,
                is_archived=False,
                organization_id=organization.id,
            ).count()
            == 1
        )


class RegistrationValidationTestCase(SeasonTestCaseMixin):
    def setUp(self):
        self.client = StateManagerAuthClient()
        super().setUp()

        self.coordinator = UserFactory()
        assign_role(self.coordinator, "coordinator")

        # Put the sessions for our Coordinator's orga
        for s in self.sessions:
            s.orga.coordinator = self.coordinator
            s.orga.save()

        # Specify one Session's moment
        self.sessions[0].day = datetime.date(2021, 1, 10)
        self.sessions[0].begin = "09:00:00"
        self.sessions[0].save()

        # Specify Season moment
        self.season.year = 2021
        self.season.month_start = 1
        self.season.n_months = 3
        self.season.save()

    def test_manager_sees_pending_registrations(self):
        organization = self.sessions[0].orga
        RegistrationFactory(
            date="2021-01-20",
            organization_id=organization.id,
            coordinator_id=organization.coordinator_id,
            day_time="08:30:00",
            classes_amount=1,
        )

        response = self.client.get(reverse("registration-validate"))
        assert response.status_code == 200
        assert str(organization) in response.rendered_content
        assert "20.01.2021" in response.rendered_content

    def test_manager_can_validate_registrations(self):
        organization = self.sessions[0].orga
        registration = RegistrationFactory(
            date="2021-01-20",
            organization_id=organization.id,
            coordinator_id=organization.coordinator_id,
            day_time="08:30:00",
            classes_amount=1,
        )

        response = self.client.post(
            reverse("registration-validate"),
            data={
                "form-organization-id": organization.pk,
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "1",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-date": "20.01.2021",
                "form-0-day_time": "08:30:00",
                "form-0-classes_amount": "1",
                "form-0-id": registration.id,
                "form-0-DELETE": "",
                "form-0-is_validated": "true",
            },
        )

        assert response.status_code == 302

        # Will raise DoesNotExist or MultipleObjectsReturned if not ONE session
        # just has been created.
        session = Session.objects.filter(
            orga_id=organization.id, day="2021-01-20", begin="08:30:00"
        ).get()

        assert Qualification.objects.filter(session_id=session.id).count() == 1

    def test_manager_cant_validate_parallel_sessions(self):
        organization = self.sessions[0].orga
        registration1 = RegistrationFactory(
            date="2021-01-20",
            organization_id=organization.id,
            coordinator_id=organization.coordinator_id,
            day_time="08:30:00",
            classes_amount=1,
        )

        registration2 = RegistrationFactory(
            date="2021-01-20",
            organization_id=organization.id,
            coordinator_id=organization.coordinator_id,
            day_time="13:30:00",
            classes_amount=1,
        )

        response = self.client.post(
            reverse("registration-validate"),
            data={
                "form-organization-id": organization.pk,
                "form-TOTAL_FORMS": "2",
                "form-INITIAL_FORMS": "2",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-date": "20.01.2021",
                "form-0-day_time": "08:30:00",
                "form-0-classes_amount": "1",
                "form-0-id": registration1.id,
                "form-0-DELETE": "",
                "form-0-is_validated": "true",
                "form-1-date": "20.01.2021",
                "form-1-day_time": "08:30:00",  # Edited to the same moment
                "form-1-classes_amount": "1",
                "form-1-id": registration2.id,
                "form-1-DELETE": "",
                "form-1-is_validated": "true",
            },
        )

        assert response.status_code == 200
        assert not Session.objects.filter(
            orga_id=organization.id, day="2021-01-20", begin="08:30:00"
        ).exists()

    def test_manager_cant_validate_sessions_over_existing(self):
        organization = self.sessions[0].orga
        registration = RegistrationFactory(
            date="2021-01-20",
            organization_id=organization.id,
            coordinator_id=organization.coordinator_id,
            day_time="08:30:00",
            classes_amount=1,
        )

        response = self.client.post(
            reverse("registration-validate"),
            data={
                "form-organization-id": organization.pk,
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "1",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-date": "10.01.2021",  # Exists
                "form-0-day_time": "08:30:00",
                "form-0-classes_amount": "1",
                "form-0-id": registration.id,
                "form-0-DELETE": "",
                "form-0-is_validated": "true",
            },
        )

        assert response.status_code == 200
        assert len(response.context["organizations"][0][1].forms[0].errors) == 1



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Didier 'OdyX' Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from simple_history.models import HistoricalRecords


class AnnualStateSetting(models.Model):
    canton = models.CharField(_("Canton"), max_length=2)
    year = models.PositiveSmallIntegerField(_("Année"))
    cost_per_bike = models.DecimalField(
        _("Coût par vélo"),
        decimal_places=2,
        max_digits=8,
        default=0,
        help_text=_("CHF"),
        validators=[MinValueValidator(0)],
    )
    cost_per_participant = models.DecimalField(
        _("Coût par participant·e·s"),
        decimal_places=2,
        max_digits=8,
        default=0,
        help_text=_("CHF"),
        validators=[MinValueValidator(0)],
    )
    history = HistoricalRecords()

    class Meta:
        verbose_name = _("Configuration cantonale par année")
        verbose_name_plural = _("Configurations cantonales par année")
        unique_together = (("canton", "year"),)

    def __str__(self):
        return gettext(
            f"{self.year}: {self.canton} vélos: {self.cost_per_bike} participants: {self.cost_per_participant}"
        )



# Generated by Django 2.2.9 on 2020-04-08 12:41

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0062_data_invoiceline_historical_session"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoiceline",
            name="historical_session",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="challenge.HistoricalSession",
            ),
        ),
    ]



# Generated by Django 2.2.9 on 2020-04-08 11:47

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0059_auto_20200305_1631"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="invoiceline",
            options={"ordering": ("session__day", "session__begin")},
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.utils.translation import gettext_lazy as _

from import_export import fields, resources

from defivelo.templatetags.dv_filters import canton_abbr

from . import ORGA_FIELDS
from .models import Organization

EXPORT_FIELDS = ["id", "created_on"] + ORGA_FIELDS


class OrganizationResource(resources.ModelResource):
    abbr = fields.Field(column_name=_("Abréviation"), attribute="abbr")
    name = fields.Field(column_name=_("Nom"), attribute="name")
    address_street = fields.Field(column_name=_("Rue"), attribute="address_street")
    address_no = fields.Field(column_name=_("N°"), attribute="address_no")
    address_additional = fields.Field(
        column_name=_("Complément d’adresse"), attribute="address_additional"
    )
    address_zip = fields.Field(column_name=_("NPA"), attribute="address_zip")
    address_city = fields.Field(column_name=_("Ville"), attribute="address_city")
    address_canton = fields.Field(
        column_name=_("Canton d’affiliation"), attribute="address_canton"
    )
    website = fields.Field(column_name=_("Site web"), attribute="website")
    coordinator_fullname = fields.Field(
        column_name=_("Coordina·teur·trice"), attribute="coordinator_fullname"
    )
    coordinator_phone = fields.Field(
        column_name=_("Téléphone"), attribute="coordinator_phone"
    )
    coordinator_natel = fields.Field(
        column_name=_("Natel"), attribute="coordinator_natel"
    )
    coordinator_email = fields.Field(
        column_name=_("Email"), attribute="coordinator_email"
    )
    status = fields.Field(column_name=_("Statut"), attribute="status_full")
    comments = fields.Field(column_name=_("Remarques"), attribute="comments")

    class Meta:
        model = Organization
        fields = EXPORT_FIELDS
        export_order = EXPORT_FIELDS

    def dehydrate_address_canton(self, field):
        return canton_abbr(field.address_canton, abbr=False, long=True)

    def dehydrate_coordinator_fullname(self, field):
        return field.coordinator.get_full_name() if field.coordinator else ""

    def dehydrate_coordinator_natel(self, field):
        return (
            field.coordinator.profile.natel
            if field.coordinator and field.coordinator.profile
            else ""
        )

    def dehydrate_coordinator_phone(self, field):
        return (
            field.coordinator.profile.phone
            if field.coordinator and field.coordinator.profile
            else ""
        )

    def dehydrate_coordinator_email(self, field):
        return field.coordinator.email if field.coordinator else ""



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from localflavor.ch.forms import CHStateSelect, CHZipCodeField

from apps.common.forms import UserAutoComplete
from apps.user import STATE_CHOICES_WITH_DEFAULT
from apps.user.models import USERSTATUS_DELETED

from .models import Organization


class OrganizationForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        cantons = kwargs.pop("cantons", None)
        super().__init__(**kwargs)
        if cantons:
            # Only permit edition within the allowed cantons
            choices = self.fields["address_canton"].choices
            choices = ((k, v) for (k, v) in choices if k in cantons)
            self.fields["address_canton"].choices = choices
            self.fields["address_canton"].initial = cantons[0]
            self.fields["address_canton"].required = True

    address_canton = forms.ChoiceField(
        label=_("Canton"),
        widget=CHStateSelect,
        choices=STATE_CHOICES_WITH_DEFAULT,
        required=False,
    )
    address_zip = CHZipCodeField(label=_("NPA"), max_length=4, required=False)

    coordinator = UserAutoComplete(
        label=_("Coordina·teur·trice"),
        queryset=(
            get_user_model().objects.exclude(profile__status=USERSTATUS_DELETED)
        ).filter(groups__name="coordinator"),
        url="user-coordinators",
        required=False,
    )

    class Meta:
        model = Organization
        fields = [
            "abbr",
            "name",
            "address_street",
            "address_no",
            "address_additional",
            "address_zip",
            "address_city",
            "address_canton",
            "website",
            "coordinator",
            "status",
            "comments",
        ]



# Generated by Django 2.2.9 on 2020-03-04 10:30

from django.contrib.auth import get_user_model
from django.db import migrations
from rolepermissions.checkers import has_role
from rolepermissions.permissions import revoke_permission
from rolepermissions.roles import assign_role


def reset_power_and_state_roles(apps, schema_editor):
    User = get_user_model()
    for user in User.objects.all():
        # Re-assigning the roles will sync the default associated permissions
        for role in ["state_manager", "power_user"]:
            if has_role(user, role):
                assign_role(user, role)


def revoke_user_permissions_for_invoices(apps, schema_editor):
    User = get_user_model()
    for user in User.objects.all():
        for perm in ["challenge_invoice_cru", "challenge_invoice_reset_to_draft"]:
            for role in ["state_manager", "power_user"]:
                if has_role(user, role):
                    revoke_permission(user, perm)
        user.save()


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0055_annualstatesetting_historicalannualstatesetting"),
    ]

    operations = [
        migrations.RunPython(
            reset_power_and_state_roles, revoke_user_permissions_for_invoices
        ),
    ]



# Generated by Django 2.2.9 on 2020-04-08 13:58

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0063_invoiceline_historical_session_not_nullable"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="invoiceline",
            options={
                "ordering": ("historical_session__day", "historical_session__begin")
            },
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Didier 'OdyX' Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date

from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.urls import reverse_lazy
from django.utils.dates import MONTHS_3
from django.utils.translation import gettext_lazy as _
from django.views.generic.dates import MonthArchiveView, YearArchiveView
from django.views.generic.edit import UpdateView

from rolepermissions.mixins import HasPermissionsMixin

from defivelo.roles import user_cantons
from defivelo.views import MenuView

from .. import timesheets_overview
from ..forms import MonthlyCantonalValidationForm
from ..models import MonthlyCantonalValidation, MonthlyCantonalValidationUrl


class ValidationsMixin(HasPermissionsMixin, MenuView):
    model = MonthlyCantonalValidation
    required_permission = "cantons_mine"
    context_object_name = "mcvs"
    date_field = "date"

    def dispatch(self, request, *args, **kwargs):
        self.year = int(kwargs.pop("year"))
        try:
            self.month = int(kwargs.pop("month"))
        except KeyError:
            self.month = None
        try:
            self.managed_cantons = user_cantons(request.user)
            self.managed_cantons.sort()
        except LookupError:
            self.managed_cantons = []
        self.canton = kwargs.get("canton", None)
        if self.canton and self.canton not in self.managed_cantons:
            # Vue individuelle, mais pas dans nos cantons
            raise PermissionDenied
        self.timesheets_statuses = timesheets_overview.timesheets_validation_status(
            year=self.year,
            month=self.month,
            cantons=([self.canton] if self.canton else self.managed_cantons),
        )
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        """
        Get the list of validations
        … but also make sure all the validations exist for that month and my cantons
        """
        kwargs = {"date__year": self.year, "canton__in": self.managed_cantons}
        if self.month:
            kwargs["date__month"] = self.month
        return (
            super().get_queryset().filter(**kwargs).order_by("canton", "date")
        ).prefetch_related("validated_urls")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Add our menu_category context
        context["menu_category"] = "finance"
        context["timesheets_statuses"] = self.timesheets_statuses
        context["mcv_urls"] = (
            MonthlyCantonalValidationUrl.objects.all().prefetch_related("translations")
        )
        return context

    def queryset_all_mcvs_in_month(self, queryset, year, month):
        """
        Using a given queryset, make sure that for a given year/month, all "my" cantons
        have created MonthlyCantonalValidation objects
        """
        # At this point, make extra sure we have all the needed ones
        with transaction.atomic():
            existing_cantons = list(
                MonthlyCantonalValidation.objects.filter(
                    date__year=year, date__month=month
                ).values_list("canton", flat=True)
            )
            missing_objects = [
                MonthlyCantonalValidation(canton=c, date=date(year, month, 1))
                for c in self.managed_cantons
                if c not in existing_cantons
            ]
            # Create the MCVs for the cantons' we're about to see (if needed)
            MonthlyCantonalValidation.objects.bulk_create(
                missing_objects, ignore_conflicts=True
            )
        return queryset.filter(canton__in=self.managed_cantons, date__year=year)


class ValidationsYearView(ValidationsMixin, YearArchiveView):
    allow_future = True
    allow_empty = True
    make_object_list = True

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Swap the array to get a canton: dict entry with 12 entries
        context["timesheets_statuses_by_canton"] = {}
        for canton in self.managed_cantons:
            context["timesheets_statuses_by_canton"][canton] = {}
            for m in MONTHS_3.keys():
                context["timesheets_statuses_by_canton"][canton][m] = context[
                    "timesheets_statuses"
                ][m].get(canton, None)
        context["months"] = MONTHS_3
        return context

    def get_queryset(self):
        """
        Make sure all the MonthlyCantonalValidation objects exist in this year, so in all months
        """
        qs = super().get_queryset()
        for month in MONTHS_3.keys():
            qs = self.queryset_all_mcvs_in_month(qs, self.year, month)
        # Return the qs, that'll get the newly created ones.
        return qs


class ValidationsMonthView(ValidationsMixin, MonthArchiveView):
    allow_future = True
    allow_empty = True

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["nothing_to_do"] = all(
            [status is None for _, status in context["timesheets_statuses"].items()]
        )
        return context

    def get_queryset(self):
        """
        Make sure all the MonthlyCantonalValidation objects exist in this month
        """
        return self.queryset_all_mcvs_in_month(
            super().get_queryset(), self.year, self.month
        )


class ValidationUpdate(ValidationsMixin, UpdateView):
    context_object_name = "mcv"
    success_message = _("Validation cantonale mise à jour")
    form_class = MonthlyCantonalValidationForm

    def get_form_kwargs(self):
        fk = super().get_form_kwargs()
        fk["validator"] = self.request.user
        fk["urls"] = MonthlyCantonalValidationUrl.objects.all()
        fk["timesheets_statuses"] = self.timesheets_statuses
        return fk

    def get_initial(self):
        initial = super().get_initial()
        initial["validated"] = self.get_object().validated
        return initial

    def get_object(self, queryset=None):
        """
        Extrait l’objet à partir de l’URL plutôt que du pk
        Permet toujours une mise à jour; en créant l’objet si nécessaire
        """
        p, _ = (
            super()
            .get_queryset()
            .get_or_create(
                canton=self.canton, defaults={"date": date(self.year, self.month, 1)}
            )
        )
        return p

    def get_success_url(self):
        return reverse_lazy(
            "salary:validations-month",
            kwargs={"year": self.object.date.year, "month": self.object.date.month},
        )



# Generated by Django 1.11.4 on 2017-08-18 14:33
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0035_auto_20170818_1626"),
    ]

    operations = [
        migrations.AlterField(
            model_name="session",
            name="superleader",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="sess_monplus",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Moniteur + / Photographe",
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0010_auto_20150903_1606"),
    ]

    operations = [
        migrations.AddField(
            model_name="session",
            name="apples",
            field=models.CharField(verbose_name="Pommes", max_length=512, blank=True),
        ),
        migrations.AddField(
            model_name="session",
            name="fallback_plan",
            field=models.CharField(
                verbose_name="Mauvais temps",
                max_length=1,
                choices=[
                    ("A", "Programme déluge"),
                    ("B", "Annulation"),
                    ("C", "Report …"),
                    ("D", "Autre …"),
                ],
                blank=True,
            ),
        ),
        migrations.AddField(
            model_name="session",
            name="helpers_place",
            field=models.CharField(
                verbose_name="Lieu rendez-vous moniteurs", max_length=512, blank=True
            ),
        ),
        migrations.AddField(
            model_name="session",
            name="helpers_time",
            field=models.TimeField(
                verbose_name="Heure rendez-vous moniteurs", null=True, blank=True
            ),
        ),
    ]



from __future__ import unicode_literals

import localflavor.generic.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="address_canton",
            field=models.CharField(max_length=2, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="address_city",
            field=models.CharField(max_length=64, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="address_no",
            field=models.CharField(max_length=8, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="address_street",
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="address_zip",
            field=models.CharField(max_length=4, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="natel",
            field=models.CharField(max_length=12, blank=True),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="iban",
            field=localflavor.generic.models.IBANField(max_length=34, blank=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-08-18 14:28

from django.contrib.auth import get_user_model
from django.db import migrations
from rolepermissions.checkers import has_role
from rolepermissions.roles import assign_role, clear_roles

from defivelo.roles import DV_AVAILABLE_ROLES


def user_set_collaborator_and_reset_roles(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    User = get_user_model()
    for user in User.objects.all():
        hadrole = False
        for rolepair in DV_AVAILABLE_ROLES:
            rolestr = rolepair[0]
            if rolestr and has_role(user, rolestr):
                clear_roles(user)
                assign_role(user, rolestr)
                hadrole = True

        if not hadrole and (user.profile.formation or user.profile.actor):
            # Give collaborator to all monitors and actors
            clear_roles(user)
            assign_role(user, "collaborator")


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0060_reset_all_roles"),
    ]

    operations = [
        migrations.RunPython(
            user_set_collaborator_and_reset_roles, migrations.RunPython.noop
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0012_userprofile_birthdate"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="bagstatus",
            field=models.PositiveSmallIntegerField(
                verbose_name="Sac Défi Vélo",
                default=0,
                choices=[(0, "---"), (10, "En prêt"), (20, "Payé")],
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("orga", "0002_address"),
    ]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="comments",
            field=models.TextField(verbose_name="Remarques", blank=True),
        ),
        migrations.AddField(
            model_name="organization",
            name="coordinator_email",
            field=models.EmailField(
                max_length=254, verbose_name="Courriel", blank=True
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="coordinator_fullname",
            field=models.CharField(
                max_length=512, verbose_name="Coordinateur", blank=True
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="coordinator_natel",
            field=models.CharField(max_length=13, verbose_name="Natel", blank=True),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0022_auto_20161004_1435"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="formation_lastdate",
            field=models.DateField(
                verbose_name="Date de la dernière formation", blank=True, null=True
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


def mulitify_actors(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    UserProfile = apps.get_model("user", "UserProfile")
    for up in UserProfile.objects.all():
        if up.actor_for:
            up.actor_for_multi.add(up.actor_for)


def demulitify_actors(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    UserProfile = apps.get_model("user", "UserProfile")
    for up in UserProfile.objects.exclude(actor_for_multi=None):
        up.actor_for = up.actor_for_multi.first()
        up.save()


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0037_auto_20170822_1510"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="actor_for_multi",
            field=models.ManyToManyField(
                blank=True,
                null=True,
                related_name="actors_multi",
                to="challenge.QualificationActivity",
                verbose_name="Intervenant",
            ),
        ),
        migrations.RunPython(mulitify_actors, demulitify_actors),
    ]
